import os
import sys
import pyzo
from pyzo import translate
from pyzo.qt import QtCore, QtGui, QtWidgets, QtPrintSupport


class PdfExport(QtWidgets.QDialog):
    """
    This class is used to export an editor to a pdf.
    The content of the editor is copied in another editor,
    and then the options chosen are applied by _updateTemporaryEditor()
    """

    def __init__(self):
        super().__init__()

        self.setWindowTitle(translate("menu dialog", "Pdf Export"))

        self.resize(1000, 600)

        self._printer = QtPrintSupport.QPrinter(
            QtPrintSupport.QPrinter.PrinterMode.HighResolution
        )
        self._printer.setColorMode(QtPrintSupport.QPrinter.ColorMode.Color)

        self._preview = QtPrintSupport.QPrintPreviewWidget(self._printer)
        self._preview.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )

        self._chkLineNumbers = QtWidgets.QCheckBox(
            "Print line number", self, checked=True
        )

        self._currentEditor = pyzo.editors.getCurrentEditor()
        self._editor = pyzo.core.editor.PyzoEditor("")

        self._sliderZoom = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
        self._sliderZoom.setMinimum(-10)
        self._sliderZoom.setMaximum(10)
        self._sliderZoom.setTickInterval(1)
        self._sliderZoom.setValue(pyzo.config.view.zoom)
        self._lblZoom = QtWidgets.QLabel()
        self._updateZoomLabel()
        self._sliderZoom.valueChanged.connect(self._updateZoomLabel)

        self._chkSyntaxHighlighting = QtWidgets.QCheckBox(
            "Enable syntax highlighting", self, checked=True
        )

        self._cmbFileName = QtWidgets.QComboBox(self)
        self._cmbFileName.addItem("Do not print the file name", 0)
        self._cmbFileName.addItem("Print with file name", 1)
        self._cmbFileName.addItem("Print with file name and absolute path", 2)
        self._cmbFileName.setCurrentIndex(1)
        self._cmbFileName.setToolTip("The title at the top of the document")

        self._cmbPageOrientation = QtWidgets.QComboBox(self)
        self._cmbPageOrientation.addItem("Portrait", 0)
        self._cmbPageOrientation.addItem("Landscape", 1)
        self._cmbPageOrientation.setToolTip("Orientation of the document")

        self._btnUpdatePreview = QtWidgets.QPushButton("Update preview", self)
        self._btnUpdatePreview.clicked.connect(self._updatePreview)
        self._btnExport = QtWidgets.QPushButton("Export")
        self._btnExport.clicked.connect(self._exportPdf)
        self._btnDone = QtWidgets.QPushButton("Done", self)
        self._btnDone.clicked.connect(self.close)

        self._mainLayout = QtWidgets.QHBoxLayout()
        self.setLayout(self._mainLayout)
        self._rightLayout = QtWidgets.QVBoxLayout()
        self._mainLayout.addWidget(self._preview)
        self._mainLayout.addLayout(self._rightLayout)

        self._optionLayout = QtWidgets.QFormLayout()
        self._rightLayout.addLayout(self._optionLayout)
        self._optionLayout.addRow(self._cmbFileName)
        self._optionLayout.addRow(self._chkLineNumbers)
        self._optionLayout.addRow(self._chkSyntaxHighlighting)
        self._optionLayout.addRow(self._lblZoom, self._sliderZoom)
        self._optionLayout.addRow(self._cmbPageOrientation)

        self._bottomLayout = QtWidgets.QHBoxLayout()
        self._rightLayout.addLayout(self._bottomLayout)
        self._bottomLayout.addStretch()
        self._bottomLayout.addWidget(self._btnUpdatePreview)
        self._bottomLayout.addWidget(self._btnExport)
        self._bottomLayout.addWidget(self._btnDone)

        self._preview.paintRequested.connect(self._editor.print)

        self._updatePreview()

    def _updateTemporaryEditor(self):
        """update the temporary editor for preview and export"""

        self._editor.setZoom(pyzo.config.view.zoom + self._sliderZoom.value())

        if self._chkSyntaxHighlighting.isChecked():
            parser = self._currentEditor.parser().name()
        else:
            parser = None
        self._editor.setParser(parser)

        self._updateOrientation()

        # Print name or filename in the editor
        headerLines = []
        if self._cmbFileName.currentIndex() == 1:
            headerLines = ["# " + self._currentEditor.name, ""]
        elif self._cmbFileName.currentIndex() == 2:
            headerLines = ["# " + self._currentEditor.filename, ""]

        editorLines = self._currentEditor.toPlainText().splitlines()

        # Print line numbers in the editor
        showLineNumbers = self._chkLineNumbers.isChecked()
        if showLineNumbers:
            numDigits = len(str(len(editorLines)))
            for i, s in enumerate(editorLines):
                editorLines[i] = str(i + 1).rjust(numDigits, "0") + "| " + s

        self._editor.clear()
        cursor = self._editor.textCursor()
        cursor.beginEditBlock()
        cursor.insertText("\n".join(headerLines + editorLines))

        # Highlight line numbers
        if showLineNumbers:
            fmt = QtGui.QTextCharFormat()
            fmt.setBackground(QtGui.QColor(240, 240, 240))
            cursor.movePosition(cursor.MoveOperation.Start, cursor.MoveMode.MoveAnchor)
            notAtLastBlock = cursor.movePosition(
                cursor.MoveOperation.NextBlock,
                cursor.MoveMode.MoveAnchor,
                len(headerLines),
            )
            while notAtLastBlock:
                cursor.movePosition(
                    cursor.MoveOperation.Right, cursor.MoveMode.KeepAnchor, numDigits
                )
                cursor.setCharFormat(fmt)
                notAtLastBlock = cursor.movePosition(
                    cursor.MoveOperation.NextBlock, cursor.MoveMode.MoveAnchor
                )

        cursor.endEditBlock()

    def _updatePreview(self):
        """Update the widget preview"""
        self._updateTemporaryEditor()
        self._preview.updatePreview()

    def _exportPdf(self):
        """Exports the code as pdf, and opens file manager"""
        options = QtWidgets.QFileDialog.Option(0)
        if not pyzo.config.advanced.useNativeFileDialogs:
            options |= QtWidgets.QFileDialog.Option.DontUseNativeDialog
        filename, selectedFilter = QtWidgets.QFileDialog.getSaveFileName(
            None, "Export PDF", os.path.expanduser("~"), "*.pdf", options=options
        )
        if not filename:
            return

        if not filename.lower().endswith(".pdf"):
            filename += ".pdf"

        self._printer.setOutputFileName(filename)
        self._updateTemporaryEditor()
        self._editor.print(self._printer)

        if self._printer.printerState() == self._printer.PrinterState.Error:
            # Notify in logger
            msg = 'could not export PDF file to "{}"'.format(filename)
            print(msg)
            # Make sure the user knows
            m = QtWidgets.QMessageBox(self)
            m.setWindowTitle("Error exporting PDF file")
            m.setText(msg)
            m.setIcon(m.Icon.Warning)
            m.exec()

    def _updateZoomLabel(self):
        self._lblZoom.setText("Zoom level: {}".format(self._sliderZoom.value()))

    def _updateOrientation(self):
        """Set document in portrait or landscape orientation"""
        index = self._cmbPageOrientation.currentIndex()
        if hasattr(self._printer, "setOrientation"):  # PySide5, PyQt5
            base = QtPrintSupport.QPrinter
            orientation = [base.Portrait, base.Landscape][index]
            self._preview.setOrientation(orientation)
            self._printer.setOrientation(orientation)
        else:  # PySide6, PyQt6
            base = QtGui.QPageLayout
            orientation = [base.Orientation.Portrait, base.Orientation.Landscape][index]

            self._preview.setOrientation(orientation)
            if sys.platform == "win32":
                self._preview.setOrientation(
                    orientation
                )  # calling this a second time is a
                # workaround, otherwise the preview in Qt6 is wrong when switching orientation
                # on Windows 10 with Qt 6.5.0

            layout = QtGui.QPageLayout()
            layout.setOrientation(orientation)
            self._printer.setPageLayout(layout)



import numpy as np
import scipy as sp

# import scipy.linalg
import time


t0 = time.time()
a = np.random.normal(size=(1600, 1600))
sp.linalg.svd(a)
print(time.time() - t0)

# For channels.py, this results in pyzo closing the connection because
# the other side seems unresponsive.



"""
Code editor extensions that change its behaviour (i.e. how it reacts to keys)
"""

from ..qt import QtGui, QtCore

Qt = QtCore.Qt

from ..misc import ce_option
from ..parsers.tokens import (
    CommentToken,
    IdentifierToken,
    NonIdentifierToken,
    StringToken,
    UnterminatedStringToken,
)
from ..parsers.python_parser import MultilineStringToken, stringLiteralPrefixes


class DeflectUpDown:
    """
    When pressing the Up key in the very first line, the cursor should move to the
    beginning of the line (resp. beginning of the indentation).
    And when pressing the Down key in the very last line, the cursor should move to the
    end of the line.

    When pressing the PageUp or PageDown key, but the cursor does not move, then
    the cursor is moved to the beginning resp. end of the block which in that cases is
    the beginning resp. end of the document.
    """

    def keyPressEvent(self, event):
        if event.modifiers() in (
            Qt.KeyboardModifier.NoModifier,
            Qt.KeyboardModifier.ShiftModifier,
        ):
            if event.key() == Qt.Key.Key_Up:
                cursor = self.textCursor()
                if (
                    cursor.blockNumber() == 0
                    and cursor.columnNumber() == cursor.positionInBlock()
                ):
                    # at first line (and not after a line wrapping)
                    HomeKey.moveHome(self, event, keepX=True)
                    return
            elif event.key() == Qt.Key.Key_Down:
                cursor = self.textCursor()
                hPos = cursor.verticalMovementX()
                if cursor.blockNumber() == self.blockCount() - 1:
                    # The cursor is at the last block, but is it also in the last line if
                    # the last block consists of multiple wrapped lines?
                    shiftDown = event.modifiers() == Qt.KeyboardModifier.ShiftModifier
                    moveMode = [cursor.MoveMode.MoveAnchor, cursor.MoveMode.KeepAnchor][
                        shiftDown
                    ]
                    if cursor.block().lineCount() == 1 or not cursor.movePosition(
                        cursor.MoveOperation.Down, moveMode
                    ):
                        # cursor is definitely at the last line
                        # --> move cursor to the end of the line/block
                        cursor.movePosition(cursor.MoveOperation.EndOfBlock, moveMode)
                        cursor.setVerticalMovementX(hPos)
                        self.setTextCursor(cursor)
                        return

            elif event.key() in (Qt.Key.Key_PageUp, Qt.Key.Key_PageDown):
                posBefore = self.textCursor().position()
                super().keyPressEvent(event)
                cursor = self.textCursor()
                posAfter = cursor.position()

                if posAfter == posBefore:
                    hPos = cursor.verticalMovementX()

                    shiftDown = event.modifiers() == Qt.KeyboardModifier.ShiftModifier
                    moveMode = [cursor.MoveMode.MoveAnchor, cursor.MoveMode.KeepAnchor][
                        shiftDown
                    ]

                    if event.key() == Qt.Key.Key_PageUp:
                        moveOp = cursor.MoveOperation.StartOfBlock
                    else:
                        moveOp = cursor.MoveOperation.EndOfBlock

                    cursor.movePosition(moveOp, moveMode)
                    cursor.setVerticalMovementX(hPos)
                    self.setTextCursor(cursor)

                return

        super().keyPressEvent(event)


class MoveLinesUpDown:
    def keyPressEvent(self, event):
        if event.key() in (Qt.Key.Key_Up, Qt.Key.Key_Down) and (
            Qt.KeyboardModifier.ControlModifier & event.modifiers()
            and Qt.KeyboardModifier.ShiftModifier & event.modifiers()
        ):
            cursor = self.textCursor()
            cursor.beginEditBlock()
            try:
                self._swaplines(cursor, event.key())
            finally:
                cursor.endEditBlock()

        else:
            super().keyPressEvent(event)

    def _swaplines(self, cursor, key):
        # Get positions of selection
        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        # Get text of selected blocks
        cursor.setPosition(start, cursor.MoveMode.MoveAnchor)
        cursor.movePosition(
            cursor.MoveOperation.StartOfBlock, cursor.MoveMode.MoveAnchor
        )
        cursor.setPosition(end, cursor.MoveMode.KeepAnchor)
        cursor.movePosition(cursor.MoveOperation.EndOfBlock, cursor.MoveMode.KeepAnchor)
        text1 = cursor.selectedText()
        cursor.removeSelectedText()
        pos1 = cursor.position()

        # Move up/down
        other = [
            cursor.MoveOperation.NextBlock,
            cursor.MoveOperation.PreviousBlock,
        ][key == Qt.Key.Key_Up]
        cursor.movePosition(other, cursor.MoveMode.MoveAnchor)

        # Select text of other block
        cursor.movePosition(
            cursor.MoveOperation.StartOfBlock, cursor.MoveMode.MoveAnchor
        )
        cursor.movePosition(cursor.MoveOperation.EndOfBlock, cursor.MoveMode.KeepAnchor)
        text2 = cursor.selectedText()
        cursor.removeSelectedText()
        pos2 = cursor.position()

        # Insert text
        cursor.insertText(text1)
        pos3 = cursor.position()

        # Move back
        if key == Qt.Key.Key_Up:
            cursor.movePosition(
                cursor.MoveOperation.NextBlock, cursor.MoveMode.MoveAnchor
            )
        else:
            cursor.setPosition(pos1, cursor.MoveMode.MoveAnchor)
            pos2 += len(text2)
            pos3 += len(text2)

        # Replace text
        cursor.insertText(text2)

        # Leave original lines selected for continued movement
        cursor.setPosition(pos2, cursor.MoveMode.MoveAnchor)
        cursor.setPosition(pos3, cursor.MoveMode.KeepAnchor)
        self.setTextCursor(cursor)


class ScrollWithUpDownKeys:
    def keyPressEvent(self, event):
        if (
            event.key() in (Qt.Key.Key_Up, Qt.Key.Key_Down)
            and Qt.KeyboardModifier.ControlModifier == event.modifiers()
        ):
            s = self.verticalScrollBar()
            # h = self.cursorRect(self.textCursor()).height()
            if event.key() == Qt.Key.Key_Up:
                s.setValue(s.value() + 1)
            else:
                s.setValue(s.value() - 1)

        else:
            super().keyPressEvent(event)


class HomeKey:
    @staticmethod
    def moveHome(editor, event, keepX=False):
        # static method to avoid code duplication

        # Prepare
        cursor = editor.textCursor()
        if keepX:
            hPos = cursor.verticalMovementX()

        shiftDown = event.modifiers() == Qt.KeyboardModifier.ShiftModifier
        moveMode = [cursor.MoveMode.MoveAnchor, cursor.MoveMode.KeepAnchor][shiftDown]
        # Get leading whitespace
        text = cursor.block().text()
        leadingWhitespace = text[: len(text) - len(text.lstrip())]
        # Get current position and move to start of whitespace
        i = cursor.positionInBlock()
        cursor.movePosition(cursor.MoveOperation.StartOfBlock, moveMode)
        cursor.movePosition(
            cursor.MoveOperation.Right, moveMode, len(leadingWhitespace)
        )
        # If we were already there, move to start of block
        if cursor.positionInBlock() == i:
            cursor.movePosition(cursor.MoveOperation.StartOfBlock, moveMode)

        if keepX:
            cursor.setVerticalMovementX(hPos)

        # Done
        editor.setTextCursor(cursor)

    def keyPressEvent(self, event):
        # Home or shift + home
        if event.key() == Qt.Key.Key_Home and event.modifiers() in (
            Qt.KeyboardModifier.NoModifier,
            Qt.KeyboardModifier.ShiftModifier,
        ):
            self.moveHome(self, event)
        else:
            super().keyPressEvent(event)


class EndKey:
    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_End and event.modifiers() in (
            Qt.KeyboardModifier.NoModifier,
            Qt.KeyboardModifier.ShiftModifier,
        ):
            # Prepare
            cursor = self.textCursor()
            shiftDown = event.modifiers() == Qt.KeyboardModifier.ShiftModifier
            moveMode = [cursor.MoveMode.MoveAnchor, cursor.MoveMode.KeepAnchor][
                shiftDown
            ]
            # Get current position and move to end of line
            i = cursor.positionInBlock()
            cursor.movePosition(cursor.MoveOperation.EndOfLine, moveMode)
            # If already at end of line, move to end of block
            if cursor.positionInBlock() == i:
                cursor.movePosition(cursor.MoveOperation.EndOfBlock, moveMode)
            # Done
            self.setTextCursor(cursor)
        else:
            super().keyPressEvent(event)


class NumpadPeriodKey:
    """
    If the numpad decimal separator key is pressed, always insert
    a period (.) even if due to localization that key is mapped to a
    comma (,). When editing code, period is the decimal separator
    independent of localization
    """

    # NOTE: THIS PLUGIN IS DISABLED, see issue #720

    def keyPressEvent(self, event):
        # Check for numpad comma
        if (
            event.key() == QtCore.Qt.Key.Key_Comma
            and event.modifiers() & QtCore.Qt.KeyboardModifier.KeypadModifier
        ):
            # Create a new QKeyEvent to substitute the original one
            event = QtGui.QKeyEvent(
                event.type(),
                QtCore.Qt.Key.Key_Period,
                event.modifiers(),
                ".",
                event.isAutoRepeat(),
                event.count(),
            )

        super().keyPressEvent(event)


class Indentation:
    def __cursorIsInLeadingWhitespace(self, cursor=None):
        """
        Checks wether the given cursor is in the leading whitespace of a block, i.e.
        before the first non-whitespace character. The cursor is not modified.
        If the cursor is not given or is None, the current textCursor is used
        """
        if cursor is None:
            cursor = self.textCursor()

        # Get the text of the current block up to the cursor
        textBeforeCursor = cursor.block().text()[: cursor.positionInBlock()]
        return (
            textBeforeCursor.lstrip() == ""
        )  # If we trim it and it is empty, it's all whitespace

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        # Tab key
        if key == Qt.Key.Key_Tab:
            if modifiers == Qt.KeyboardModifier.NoModifier:
                if (
                    self.textCursor().hasSelection()
                ):  # Tab pressed while some area was selected
                    self.indentSelection()
                    return
                elif self.__cursorIsInLeadingWhitespace():
                    # If the cursor is in the leading whitespace, indent and move cursor to end of whitespace
                    cursor = self.textCursor()
                    self.indentBlock(cursor)
                    self.setTextCursor(cursor)
                    return

                elif self.indentUsingSpaces():
                    # Insert space-tabs
                    cursor = self.textCursor()
                    w = self.indentWidth()
                    cursor.insertText(" " * (w - ((cursor.positionInBlock() + w) % w)))
                    return

        # If backspace is pressed in the leading whitespace, (except for at the first
        # position of the line), and there is no selection
        # dedent that line and move cursor to end of whitespace
        elif (
            key == Qt.Key.Key_Backspace
            and modifiers == Qt.KeyboardModifier.NoModifier
            and self.__cursorIsInLeadingWhitespace()
            and not self.textCursor().atBlockStart()
            and not self.textCursor().hasSelection()
        ):
            # Create a cursor, dedent the block and move screen cursor to the end of the whitespace
            cursor = self.textCursor()
            self.dedentBlock(cursor)
            self.setTextCursor(cursor)
            return

        # todo: Same for delete, I think not (what to do with the cursor?)

        # Auto-unindent
        elif key == Qt.Key.Key_Delete:
            cursor = self.textCursor()
            if not cursor.hasSelection():
                cursor.movePosition(
                    cursor.MoveOperation.EndOfBlock, cursor.MoveMode.KeepAnchor
                )
                if not cursor.hasSelection() and cursor.block().next().isValid():
                    cursor.beginEditBlock()
                    cursor.movePosition(cursor.MoveOperation.NextBlock)
                    self.indentBlock(cursor, -99)  # dedent as much as we can
                    cursor.deletePreviousChar()
                    cursor.endEditBlock()
                    return

        super().keyPressEvent(event)


class AutoIndent:
    """
    Auto indentation. This extension only adds the autoIndent property, for the
    actual indentation, the editor should derive from some AutoIndenter object
    """

    def autoIndent(self):
        """Get whether auto indentation is enabled."""
        return self.__autoIndent

    @ce_option(True)
    def setAutoIndent(self, value):
        """Set whether to enable auto indentation."""
        self.__autoIndent = bool(value)


class PythonAutoIndent:
    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if not self.autoIndent():
            return

        # This extension code is run *after* key is processed by QPlainTextEdit

        if event.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
            cursor = self.textCursor()

            # Prevent in-block newlines (issue #482)
            if not cursor.atBlockStart() and not cursor.hasSelection():
                cursor.deletePreviousChar()
                cursor.insertBlock()
                cursor = self.textCursor()

            previousBlock = cursor.block().previous()
            if previousBlock.isValid():
                line = previousBlock.text()
                indent = line[: len(line) - len(line.lstrip())]
                if ":" in line:
                    # We only need to add indent if the : is not in a (multiline)
                    # string or comment. Therefore, find out what the syntax
                    # highlighter thinks of the previous line.
                    ppreviousBlock = (
                        previousBlock.previous()
                    )  # the block before previous
                    ppreviousState = (
                        ppreviousBlock.userState() if previousBlock.isValid() else 0
                    )
                    tokens = self.parser().parseLine(
                        previousBlock.text(), ppreviousState
                    )
                    # because of the ":" on that line, there is at least one token
                    t = tokens[-1]
                    if isinstance(t, CommentToken):
                        t = tokens[-2] if len(tokens) >= 2 else None
                    if isinstance(t, NonIdentifierToken) and str(t).strip() == ":":
                        # there is no need to check if the previous indent is the same
                        # style as the added one, because we could not fix that anyways
                        if self.indentUsingSpaces():
                            indent += " " * self.indentWidth()
                        else:
                            indent += "\t"
                cursor.insertText(indent)
                # This prevents jump to start of line when up key is pressed
                self.setTextCursor(cursor)


class SmartCopyAndPaste:
    """
    Smart copy and paste allows copying and pasting blocks

    """

    @staticmethod
    def __setCursorPositionAndAnchor(cursor, position, anchor):
        cursor.setPosition(anchor)
        cursor.setPosition(position, cursor.MoveMode.KeepAnchor)

    @classmethod
    def __ensureCursorBeforeAnchor(cls, cursor):
        """
        Given a cursor, modify it such that the cursor.position() is before or
        at cursor.anchor() and not the other way around.

        Returns: anchorBeforeCursor, i.e. whether originally the anchor was
        before the cursor
        """
        start = cursor.selectionStart()
        end = cursor.selectionEnd()
        # Remember whether the cursor is before or after the anchor
        anchorBeforeCursor = cursor.anchor() < cursor.position()

        cls.__setCursorPositionAndAnchor(cursor, start, end)

        # Return wheter originally the cursor was before the anchor
        return anchorBeforeCursor

    def copy(self):
        """Smart copy: if selection is multi-line and in front of the start of the
        selection there is only whitespace, extend the selection to include only
        whitespace
        """
        cursor = self.textCursor()
        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        # For our convenience, ensure that position is at start and
        # anchor is at the end, but remember whether originally the
        # anchor was before the cursor or the other way around
        anchorBeforeCursor = self.__ensureCursorBeforeAnchor(cursor)

        # Check if we have multi-line selection.
        block = cursor.block()
        # Use > not >= to ensure we don't count it as multi-line if the cursor
        # is just at the beginning of the next block (consistent with 'CodeEditor.doForSelectedLines')
        if end > (block.position() + block.length()):
            # Now check if there is only whitespace before the start of selection
            # If so, include this whitespace in the selection and update the
            # selection of the editor
            textBeforeSelection = block.text()[: cursor.positionInBlock()]
            if len(textBeforeSelection.strip()) == 0:
                start = block.position()  # Move start to include leading whitespace

                # Update the textcursor of our editor. If originally the
                # anchor was before the cursor, restore that situation
                if anchorBeforeCursor:
                    self.__setCursorPositionAndAnchor(cursor, end, start)
                else:
                    self.__setCursorPositionAndAnchor(cursor, start, end)

                # Setting the textcursor might scroll to another place
                # (e.g. Ctrl+A, Ctrl+C would scroll to the very bottom).
                # To prevent this, we store and then restore the position
                # of the vertical scrollbar
                sb = self.verticalScrollBar()
                scrollbarPos = sb.value()
                self.setTextCursor(cursor)
                sb.setValue(scrollbarPos)

        # Call our supers copy slot to do the actual copying
        super().copy()

    def cut(self):
        """Cutting with smart-copy: the part that is copied is the same as self.copy(),
        but the part that is removed is only the original selection

        see: Qt qtextcontrol.cpp, cut()
        """
        if (
            self.textInteractionFlags() & QtCore.Qt.TextInteractionFlag.TextEditable
        ) and self.textCursor().hasSelection():
            cursor = self.textCursor()
            self.copy()
            # Restore original cursor
            self.setTextCursor(cursor)
            cursor.removeSelectedText()

    def paste(self):
        """Smart paste
        If you paste on a position that has only whitespace in front of it,
        remove the whitespace before pasting. Combined with smart copy,
        this ensure indentation of the
        """
        self._paste(keepSelection=False)

    def pasteAndSelect(self):
        """Smart paste
        Like paste(), but keep the part that was pasted selected. This allows
        you to change the indentation after pasting using tab / shift-tab
        """
        self._paste(keepSelection=True)

    def _paste(self, keepSelection):
        # Create a cursor of the current selection
        cursor = self.textCursor()

        # Ensure that position is at start and anchor is at the end.
        # This is required to ensure that with keepPositionOnInsert
        # set, the cursor will equal the pasted text after the pasting
        self.__ensureCursorBeforeAnchor(cursor)

        # Change this cursor to let the position() stay at its place upon
        # inserting the new code; the anchor will move to the end of the insertion
        cursor.setKeepPositionOnInsert(True)
        super().paste()

        block = cursor.block()

        # Check if the thing to be pasted is multi-line. Use > not >=
        # to ensure we don't count it as multi-line if the cursor
        # is just at the beginning of the next block (consistent with
        # 'CodeEditor.doForSelectedLines')
        if cursor.selectionEnd() > block.position() + block.length():
            # Now, check if in front of the current selection there is only whitespace
            if len(block.text()[: cursor.positionInBlock()].strip()) == 0:
                # Note that this 'smart pasting' will be a separate item on the
                # undo stack. This is intentional: the user can undo the 'smart paste'
                # without undoing the paste
                cursor2 = QtGui.QTextCursor(cursor)
                cursor2.setPosition(
                    cursor2.position()
                )  # put the anchor where the cursor is

                # Move cursor2 to beginning of the line (selecting the whitespace)
                # and remove the selection
                cursor2.movePosition(
                    cursor2.MoveOperation.StartOfBlock, cursor2.MoveMode.KeepAnchor
                )
                cursor2.removeSelectedText()

        # Set the textcursor of this editor to the just-pasted selection
        if keepSelection:
            cursor.setKeepPositionOnInsert(False)
            self.setTextCursor(cursor)


class AutoCloseQuotesAndBrackets:
    """
    Automatic insertion of quotes, parenthesis, braces and brackets

    """

    def autoClose_Brackets(self):
        return self.__autoClose_Brackets

    @ce_option(False)
    def setAutoClose_Brackets(self, value):
        self.__autoClose_Brackets = value

    def autoClose_Quotes(self):
        return self.__autoClose_Quotes

    @ce_option(False)
    def setAutoClose_Quotes(self, value):
        self.__autoClose_Quotes = value

    def _get_token_at_cursor(self, cursor=None, relpos=0):
        """Get token at the (current or given) cursor position. Can be None."""
        if cursor is None:
            cursor = self.textCursor()
        pos = cursor.positionInBlock() + relpos
        tokenAtCursor = None
        for token in cursor.block().userData().tokens:
            if hasattr(token, "start"):
                if token.start > pos:
                    break
                if pos < token.end or isinstance(token, CommentToken):
                    # When the cursor is at the end of a comment line,
                    # then "pos" is equal to "token.end", but we are still inside the
                    # comment token.
                    tokenAtCursor = token
                    break
            if isinstance(token, MultilineStringToken):
                # multiline string tokens can have a length
                # of zero if the whole line is an empty comment
                if token.start == token.end == 0:
                    tokenAtCursor = token
                    break
        return tokenAtCursor

    def keyPressEvent(self, event):
        try:
            self.__keyPressEvent(event)
        except Exception as err:
            # When there is a bug in our fancy autoclosing stuff, better print and
            # and have the plain behavior instead of not working ...
            print(err)
            super().keyPressEvent(event)

    def __keyPressEvent(self, event):
        quotes = "'", '"'
        openBrackets = "{", "[", "("
        closeBrackets = "}", "]", ")"
        brackets = openBrackets + closeBrackets

        cursor = self.textCursor()
        char = event.text()

        #  brackets
        if char in brackets and self.__autoClose_Brackets:
            # Don't autobracket inside comments and strings
            if isinstance(
                self._get_token_at_cursor(cursor),
                (
                    CommentToken,
                    StringToken,
                    MultilineStringToken,
                    UnterminatedStringToken,
                ),
            ):
                super().keyPressEvent(event)
                return

            if char in openBrackets:
                idx = openBrackets.index(char)
                next_char = self.__getNextCharacter()
                if cursor.selectedText():
                    # Surround selection with brackets
                    new_text = (
                        openBrackets[idx] + cursor.selectedText() + closeBrackets[idx]
                    )
                    cursor.setKeepPositionOnInsert(True)
                    cursor.insertText(new_text)
                    cursor.setKeepPositionOnInsert(False)
                    self.setTextCursor(cursor)
                elif (
                    next_char.strip() and next_char not in closeBrackets
                ):  # str.strip() conveniently removes all kinds of whitespace
                    # Only autoclose if the char on the right is whitespace
                    cursor.insertText(char)  # == super().keyPressEvent(event)
                else:
                    # Auto-close bracket
                    insert_txt = "{}{}".format(openBrackets[idx], closeBrackets[idx])
                    cursor.insertText(insert_txt)
                    self._moveCursorLeft(1)

            elif char in closeBrackets:
                idx = closeBrackets.index(char)
                next_char = self.__getNextCharacter()
                if cursor.selectedText():
                    # Replace
                    cursor.insertText(char)
                elif next_char == char:
                    # Skip
                    self._moveCursorRight(1)
                else:
                    # Normal
                    cursor.insertText(char)  # == super().keyPressEvent(event)

        # quotes
        elif char in quotes and self.__autoClose_Quotes:
            next_char = self.__getNextCharacter()

            # Don't autoquote inside comments and multiline strings
            # Only allow "doing our thing" when we're at the end of a normal string
            token = self._get_token_at_cursor(cursor)
            if isinstance(token, (CommentToken, MultilineStringToken)):
                super().keyPressEvent(event)
                return
            elif isinstance(token, StringToken) and char != next_char:
                super().keyPressEvent(event)
                return

            if cursor.selectedText() in ('"', "'"):
                # Skip over char if it's one char and a quote
                self._moveCursorRight(1)
            elif cursor.selectedText():
                # Surround selection with quotes, maybe even multi-line
                new_text = char + cursor.selectedText() + char
                if "\u2029" in new_text and "python" in self.parser().name().lower():
                    new_text = char * 2 + new_text + char * 2
                cursor.setKeepPositionOnInsert(True)
                cursor.insertText(new_text)
                cursor.setKeepPositionOnInsert(False)
                self.setTextCursor(cursor)
            elif next_char == char:
                # Skip
                self._moveCursorRight(1)
            else:
                # Only autoquote if we're next to whitespace, operator, quote
                tokenL = self._get_token_at_cursor(cursor, -1)
                tokenR = self._get_token_at_cursor(cursor, -0)

                if (
                    isinstance(tokenR, IdentifierToken)
                    or isinstance(
                        tokenL, StringToken
                    )  # don't add a fourth " when typing " after two "
                    or (
                        isinstance(tokenL, IdentifierToken)
                        and str(tokenL).lower() not in stringLiteralPrefixes
                    )
                ):
                    super().keyPressEvent(event)
                    return
                # Auto-close
                cursor.insertText(char * 2)
                self._moveCursorLeft(1)
                # # Maybe handle triple quotes (add 2 more if we now have 3 on the left)
                # Disabled: this feature too easily gets in the way
                # if 'python' in self.parser().name().lower():
                #     cursor = self.textCursor()
                #     cursor.movePosition(cursor.MoveOperation.PreviousCharacter, cursor.MoveMode.KeepAnchor, 3)
                #     if cursor.selectedText() == char * 3:
                #         edit_cursor = self.textCursor()
                #         edit_cursor.insertText(char * 2)
                #         self._moveCursorLeft(2)

        # remove whole couple of brackets when hitting backspace
        elif event.key() == Qt.Key.Key_Backspace and self.__autoClose_Brackets:
            if isinstance(
                self._get_token_at_cursor(cursor),
                (
                    CommentToken,
                    StringToken,
                    MultilineStringToken,
                    UnterminatedStringToken,
                ),
            ):
                super().keyPressEvent(event)
                return

            else:
                prev_char = self.__getPreviousCharacter()
                next_char = self.__getNextCharacter()

                if prev_char == "(" and next_char == ")":
                    cursor.deleteChar()
                elif prev_char == "[" and next_char == "]":
                    cursor.deleteChar()
                elif prev_char == "{" and next_char == "}":
                    cursor.deleteChar()

                super().keyPressEvent(event)

        else:
            super().keyPressEvent(event)

    def __getNextCharacter(self):
        cursor = self.textCursor()
        cursor.movePosition(
            cursor.MoveOperation.NoMove, cursor.MoveMode.MoveAnchor
        )  # rid selection
        cursor.movePosition(
            cursor.MoveOperation.NextCharacter, cursor.MoveMode.KeepAnchor
        )
        next_char = cursor.selectedText()
        return next_char

    def __getPreviousCharacter(self):
        cursor = self.textCursor()
        cursor.movePosition(
            cursor.MoveOperation.NoMove, cursor.MoveMode.MoveAnchor
        )  # rid selection
        cursor.movePosition(
            cursor.MoveOperation.PreviousCharacter, cursor.MoveMode.KeepAnchor
        )
        previous_char = cursor.selectedText()
        return previous_char

    def _moveCursorLeft(self, n):
        """Move cursor left between eg. brackets"""
        cursor2 = self.textCursor()
        cursor2.movePosition(cursor2.MoveOperation.Left, cursor2.MoveMode.MoveAnchor, n)
        self.setTextCursor(cursor2)

    def _moveCursorRight(self, n):
        """Move cursor out of eg. brackets"""
        cursor2 = self.textCursor()
        cursor2.movePosition(
            cursor2.MoveOperation.Right, cursor2.MoveMode.MoveAnchor, n
        )
        self.setTextCursor(cursor2)



import inspect
import os

__this_dir__ = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))
os.chdir(__this_dir__)

##


def make_path_rounded_rect(x_left, y_top, w, h, r):
    return ' '.join([
        f'M {x_left + r} {y_top}',  # move to x y
        f'L {x_left + w - r} {y_top}',  # draw line to x y
        f'A {r} {r} 0 0 1 {x_left + w} {y_top + r}',
        f'L {x_left + w} {y_top + h - r}',  # draw line to x y
        f'A {r} {r} 0 0 1 {x_left + w - r} {y_top + h}',
        f'L {x_left + r} {y_top + h}',  # draw line to x y
        f'A {r} {r} 0 0 1 {x_left} {y_top + h - r}',
        f'L {x_left} {y_top + r}',  # draw line to x y
        f'A {r} {r} 0 0 1 {x_left + r} {y_top}',
        'z',  # close path
    ])

def make_path_gtsign(x_inner, y_inner, thickness, r_outer, length_inner):
    r = r_outer
    r2 = 0.5 * thickness
    k = 1 / 2**0.5
    x1 = x_inner - k * length_inner
    y1 = y_inner - k * length_inner
    x2 = x1 + k * 2 * r2
    y2 = y1 - k * 2 * r2
    x3 = x_inner + thickness / k - k * r
    y3 = y_inner - k * r

    return ' '.join([
        f'M {x_inner} {y_inner}',  # move to x y
        f'L {x1} {y1}',  # draw line to x y
        f'A {r2} {r2} 0 0 1 {x2} {y2}',
        f'L {x3} {y3}',  # draw line to x y
        f'A {r} {r} 0 0 1 {x3} {y_inner + (y_inner - y3)}',
        f'L {x2} {y_inner + (y_inner - y2)}',  # draw line to x y
        f'A {r2} {r2} 0 0 1 {x1} {y_inner + (y_inner - y1)}',
        f'L {x_inner} {y_inner}',  # draw line to x y
        'z',  # close path
    ])

def make_path_textline(x, y, thickness, length):
    r = 0.5 * thickness
    return ' '.join([
        f'M {x + r} {y - r}',  # move to x y
        f'L {x + length - r} {y - r}',  # draw line to x y
        f'A {r} {r} 0 0 1 {x + length - r} {y + r}',
        f'L {x + r} {y + r}',  # draw line to x y
        f'A {r} {r} 0 0 1 {x + r} {y - r}',
        'z',  # close path
    ])


def make_pyzologo(target_size=256):
    svg_contents = []
    svg_contents.append('<?xml version="1.0" encoding="UTF-8" standalone="no"?>')
    svg_contents.append(f'<svg width="{target_size}" height="{target_size}" viewBox="0 0 {target_size} {target_size}" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">')
    svg_contents.append(f'<!-- created via a Python script - do not edit this SVG file directly -->')

    # add filled background, only for debugging
    # svg_contents.append(f'<rect fill="#777" stroke="#000" x="0" y="0" width="{target_size}" height="{target_size}"/>')

    # https://developer.mozilla.org/en-US/docs/Web/SVG/Tutorials/SVG_from_scratch/Paths

    def scale_args(*args):
        return [v / 256 * target_size for v in args]

    for shadow in [True, False]:
        sx = sy = 0
        sc = ''
        if shadow:
            sx = sy = 4
            sc = '#' + 'b0' * 3  # fixed color shadow, without transparency

        svg_contents.append('<path d="{}" fill="{}" />'.format(make_path_rounded_rect(*scale_args(12 + sx, 45 + sy, 104, 168, 8)), sc or '#268bd2'))
        svg_contents.append('<path d="{}" fill="{}" />'.format(make_path_rounded_rect(*scale_args(124 + sx, 45 + sy, 120, 88, 8)), sc or '#dc322f'))
        svg_contents.append('<path d="{}" fill="{}" />'.format(make_path_rounded_rect(*scale_args(124 + sx, 141 + sy, 56, 72, 8)), sc or '#859900'))
        svg_contents.append('<path d="{}" fill="{}" />'.format(make_path_rounded_rect(*scale_args(188 + sx, 141 + sy, 56, 72, 8)), sc or '#859900'))

    for i in range(3):
        svg_contents.append('<path d="{}" fill="#a52523" />'.format(make_path_gtsign(*scale_args(154 + i * 32, 80, 8, 3, 19))))

    for i in range(4):
        svg_contents.append('<path d="{}" fill="#1c689d" />'.format(make_path_textline(*scale_args(27, 80 + i * 32, 8, 73 if i != 3 else 57))))

    svg_contents.append('</svg>')

    return svg_contents


if __name__ == '__main__':
    target_size = 256
    svg_contents = make_pyzologo(target_size)
    with open('pyzologo{}.svg'.format(target_size), 'wt') as fd:
        fd.write('\n'.join(svg_contents))




"""
Module to integrate GUI event loops in the Pyzo interpreter.

This specifies classes that all have the same interface. Each class
wraps one GUI toolkit.

Support for PySide2/6, PyQt5/6, WxPython, FLTK, GTK, TK.

"""

import sys
import time
import importlib
import functools

from pyzokernel import printDirect


# Warning message.
mainloopWarning = (
    """
Note: The GUI event loop is already running in the pyzo kernel. Be aware
that the function to enter the main loop does not block.
""".strip()
    + "\n"
)

# Qt has its own message
mainloopWarning_qt = """
Note on using QApplication.exec_():
The GUI event loop is already running in the pyzo kernel, and exec_()
does not block. In most cases your app should run fine without the need
for modifications. For clarity, this is what the pyzo kernel does:
- Prevent deletion of objects in the local scope of functions leading to exec_()
- Prevent system exit right after the exec_() call
""".lstrip()


# Print the main loop warning at most once
_printed_warning = False


def print_mainloop_warning(msg=None):
    global _printed_warning
    if not _printed_warning:
        _printed_warning = True
        msg = msg or mainloopWarning
        printDirect(msg)


class App_base:
    """Defines the interface."""

    def process_events(self):
        pass

    def _keyboard_interrupt(self, signum=None, frame=None):
        interpreter = sys._pyzoInterpreter
        interpreter.write("\nKeyboardInterrupt\n")
        interpreter._resetbuffer()
        if interpreter.more:
            interpreter.more = 0
            interpreter.newPrompt = True

    def run(self, repl_callback, sleep_time=0.01):
        """Very simple mainloop. Subclasses can overload this to use
        the native event loop. Attempt to process GUI events every so often.
        """

        # The sleep_time is 100Hz by default, which seems like an ok balance
        # between CPU strain and smooth animations. Ideally we'd run a
        # real event loop though, that is fast when needed and just sleeps
        # when the gui is idle, saving battery life.

        perf_counter = getattr(time, "perf_counter", time.time)

        repl_time = 0.099
        next_repl = perf_counter() + repl_time

        # The toplevel while-loop is just to catch Keyboard interrupts
        # and then proceed. The inner while-loop is the actual event loop.
        while True:
            try:
                while True:
                    time.sleep(sleep_time)
                    self.process_events()
                    if perf_counter() > next_repl:
                        next_repl = perf_counter() + repl_time
                        repl_callback()

            except KeyboardInterrupt:
                self._keyboard_interrupt()
            except TypeError:
                # For some reason, when wx is integrated, keyboard interrupts
                # result in a TypeError.
                # I tried to find the source, but did not find it. If anyone
                # has an idea, please e-mail me!
                if "_wx" in self.__class__.__name__.lower():
                    self._keyboard_interrupt()

    def quit(self):
        raise SystemExit()


class App_nogui(App_base):
    """The app for when there is no GUI."""

    def run(self, repl_callback):
        # Move at a slow pace - there is no gui to keep running
        App_base.run(self, repl_callback, 0.1)


# Experimental and WIP
class App_asyncio_new(App_base):
    """Based on asyncio (standard Python) event loop.

    Actually run an event loop and support switching to another loop.
    """

    def __init__(self):
        import asyncio

        asyncio.integrate_with_ide = self.enable
        self._loop = None

    def run(self, repl_callback):
        import asyncio

        self._repl_callback = repl_callback
        self._sleep_time = 0.1

        if sys.version_info < (3, 12):
            loop = asyncio.get_event_loop_policy().get_event_loop()
        else:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                # no running event loop
                loop = asyncio.new_event_loop()
        self.enable(loop, True)

    def enable(self, loop, run=False):
        if loop is not self._loop:
            self.swap_loops_when_new_one_starts(self._loop, loop)
            loop.call_later(self._sleep_time, self._ping_repl_callback)
        if run:
            loop.run_forever()

    def swap_loops_when_new_one_starts(self, old_loop, new_loop):
        def new_run_forever(*args, **kwargs):
            if old_loop and old_loop.is_running():
                old_loop.stop()
            self._loop = new_loop
            if not new_loop.is_running():
                while True:
                    try:
                        new_loop.original_run_forever(*args, **kwargs)
                        break
                    except KeyboardInterrupt:
                        self._keyboard_interrupt()
                        continue

        if not hasattr(new_loop, "original_run_forever"):
            new_loop.original_run_forever = new_loop.run_forever
        new_loop.run_forever = new_run_forever

    def _ping_repl_callback(self):
        import asyncio

        self._repl_callback()

        # Get loop that (probably) called this
        try:
            loop = asyncio.get_running_loop()
        except Exception:
            loop = None

        # If it's the same as our current loop, we want to be called again
        if loop:
            self._loop.call_later(self._sleep_time, self._ping_repl_callback)

    def quit(self):
        if self._loop:
            self._loop.stop()
        # raise SystemExit()


class App_asyncio(App_base):
    """Based on asyncio (standard Python) event loop.

    We do not run the event loop and a timer to keep the REPL active, because
    asyncio does allow creating new loops while one is running. So we stick to
    a simple and non-intrusive mechanism to regularly process events from
    whatever event loop is current at any given moment.

    """

    def __init__(self):
        import asyncio

        self.app = asyncio.get_event_loop_policy().get_event_loop()
        self.app._in_event_loop = "Pyzo"
        self._warned_about_process_events = False
        self._blocking = False

        # Hijack

        # Prevent entering forever, not giving control back to repl
        self.app._original_run_forever = self.app.run_forever
        self.app.run_forever = self.stub_run_forever
        # The run_until_complete() calls run_forever and then checks that future completed
        self.app._original_run_until_complete = self.app.run_until_complete
        self.app.run_until_complete = self.stub_run_until_complete
        # Prevent the loop from being destroyed
        self.app._original_close = self.app.close
        self.app.close = self.stub_close
        # Stop is fine, since we don't "run" the loop, but just keep it active

    def stub_run_until_complete(self, *args):
        self._blocking = True
        return self.app._original_run_until_complete(*args)

    def stub_run_forever(self):
        if self._blocking:
            self._blocking = False
            self.app._original_run_forever()

    def stub_close(self):
        pass

    def process_events(self):
        import asyncio

        try:
            loop = asyncio.get_event_loop_policy().get_event_loop()
        except Exception:
            loop = None
        if loop is not self.app:
            return  # The loop was replaced
        loop = self.app
        if loop.is_closed():
            pass  # not much we can do
        elif loop.is_running():
            if not self._warned_about_process_events:
                print("Warning: cannot process events synchronously in asyncio")
                self._warned_about_process_events = True
        else:
            # First calling stop and then run_forever() process all pending events.
            # We do this multiple times to work around the limited frequence that this
            # method gets called, but we need to use a private attribute for that :/
            for i in range(20):
                loop.stop()
                loop._original_run_forever()
                if len(getattr(loop, "_ready", ())) == 0:
                    break

    def quit(self):
        loop = self.app
        try:
            if not loop.is_closed():
                loop.close()
        finally:
            raise SystemExit()


class App_tk(App_base):
    """Replaces tkinter's mainloop with a dummy so when a module calls mainloop, it does
    not block.

    Instead of the blocking mainloop, tk events are processed by calling the "update"
    method of the tk root object.
    In order to obtain the (newest) tk root object, the "__init__" method of the "Tk"
    class is wrapped by our function "hijacked_tk_init".

    If the user executes "root = tkinter.Tk()" multiple times, older root objects will
    be destroyed by Pyzo. This way it is more comfortable, e.g. wenn re-executing the
    code of a file or of a code cell.
    """

    def __init__(self):
        if sys.version_info[0] == 3:
            import tkinter
        else:
            import Tkinter as tkinter

        self._original_tk_init = tkinter.Tk.__init__
        self._tk_root = None

        def hijacked_tk_init(tk_self, *args, **kwargs):
            if self._tk_root is not None:
                # tkinter.Tk() is already called the second time or more often
                try:
                    # this might fail if _tk_root was already destroyed
                    self._tk_root.destroy()
                    print("Note: Pyzo GUI integration destroyed the old Tk root.")
                except Exception:
                    pass
            self._tk_root = tk_self

            # Notify that we integrated the event loop
            self._tk_root._in_event_loop = "Pyzo"
            tkinter._in_event_loop = "Pyzo"

            return self._original_tk_init(tk_self, *args, **kwargs)

        # update docstring etc.
        tkinter.Tk.__init__ = functools.wraps(self._original_tk_init)(hijacked_tk_init)

        # Replace mainloop. Note that a root object obtained with
        # tkinter.Tk() has a mainloop method, which will simply call
        # tkinter.mainloop().
        def dummy_mainloop(*args, **kwargs):
            print_mainloop_warning()

        tkinter.Misc.mainloop = dummy_mainloop
        tkinter.mainloop = dummy_mainloop

        self._tkinter = tkinter

    def process_events(self):
        if self._tk_root is not None:
            # In older tkinter versions (e.g. the one of Python 2.7), calling "update"
            # on a destroyed tk root object throws an exception which would terminate
            # the shell if not caught.
            try:
                self._tk_root.update()
            except Exception:
                pass


class App_fltk(App_base):
    """Hijack fltk 1.
    This one is easy. Just call fl.wait(0.0) now and then.
    Note that both tk and fltk try to bind to PyOS_InputHook. Fltk
    will warn about not being able to and Tk does not, so we should
    just hijack (import) fltk first. The hook that they try to fetch
    is not required in pyzo, because the pyzo interpreter will keep
    all GUI backends updated when idle.
    """

    def __init__(self):
        # Try importing
        import fltk as fl
        import types

        # Replace mainloop with a dummy
        def dummyrun(*args, **kwargs):
            print_mainloop_warning()

        fl.Fl.run = types.MethodType(dummyrun, fl.Fl)

        # Store the app instance to process events
        self.app = fl.Fl

        # Notify that we integrated the event loop
        self.app._in_event_loop = "Pyzo"
        fl._in_event_loop = "Pyzo"

    def process_events(self):
        self.app.wait(0)


class App_fltk2(App_base):
    """Hijack fltk 2."""

    def __init__(self):
        # Try importing
        import fltk2 as fl

        # Replace mainloop with a dummy
        def dummyrun(*args, **kwargs):
            print_mainloop_warning()

        fl.run = dummyrun

        # Return the app instance to process events
        self.app = fl

        # Notify that we integrated the event loop
        self.app._in_event_loop = "Pyzo"

    def process_events(self):
        # is this right?
        self.app.wait(0)


class App_tornado(App_base):
    """Hijack Tornado event loop.

    Tornado does have a function to process events, but it does not
    work when the event loop is already running. Therefore we don't
    enter the real Tornado event loop, but just poll it regularly.
    """

    def __init__(self):
        # Try importing
        import tornado.ioloop

        # Get the "app" instance
        self.app = tornado.ioloop.IOLoop.instance()

        # Replace mainloop with a dummy
        def dummy_start():
            print_mainloop_warning()
            sys._pyzoInterpreter.ignore_sys_exit = True
            self.app.add_callback(reset_sys_exit)

        def dummy_stop():
            pass

        def reset_sys_exit():
            sys._pyzoInterpreter.ignore_sys_exit = False

        def run_sync(func, timeout=None):
            self.app.start = self.app._original_start
            try:
                self.app._original_run_sync(func, timeout)
            finally:
                self.app.start = self.app._dummy_start

        #
        self.app._original_start = self.app.start
        self.app._dummy_start = dummy_start
        self.app.start = self.app._dummy_start
        #
        self.app._original_stop = self.app.stop
        self.app._dummy_stop = dummy_stop
        self.app.stop = self.app._dummy_stop
        #
        self.app._original_run_sync = self.app.run_sync
        self.app.run_sync = run_sync

        # Notify that we integrated the event loop
        self.app._in_event_loop = "Pyzo"

        self._warned_about_process_events = False

    def process_events(self):
        if not self._warned_about_process_events:
            print("Warning: cannot process events synchronously in Tornado")
            self._warned_about_process_events = True
        # self.app.run_sync(lambda x=None: None)

    def run(self, repl_callback):
        from tornado.ioloop import PeriodicCallback

        # Create timer
        self._timer = PeriodicCallback(repl_callback, 0.05 * 1000)
        self._timer.start()
        # Enter mainloop
        self.app._original_start()
        while True:
            try:
                self.app._original_start()
            except KeyboardInterrupt:
                self._keyboard_interrupt()
                self.app._original_stop()
                continue
            break

    def quit(self):
        self.app._original_stop()
        # raise SystemExit()


class App_qt(App_base):
    """Common functionality for pyqt and pyside"""

    _API = None  # this will be overridden in the derived class

    def __init__(self):
        import types

        # Try importing qt
        QtCore = importlib.import_module(self._API + ".QtCore")
        # QtGui = importlib.import_module(self._API + ".QtGui")
        QtWidgets = importlib.import_module(self._API + ".QtWidgets")
        self._QtWidgets, self._QtCore = QtWidgets, QtCore

        # Store the real application class
        if not hasattr(QtWidgets, "real_QApplication"):
            QtWidgets.real_QApplication = QtWidgets.QApplication

        class QApplication_hijacked(QtWidgets.QApplication):
            """QApplication_hijacked(*args, **kwargs)

            Hijacked QApplication class. This class has a __new__()
            method that always returns the global application
            instance, i.e. QtWidgets.qApp.

            The QtWidgets.qApp instance is an instance of the original
            QtWidgets.QApplication, but with its __init__() and exec_()
            methods replaced.

            You can subclass this class; the global application instance
            will be given the methods and attributes so it will behave
            like the subclass.
            """

            def __new__(cls, *args, **kwargs):
                # Get the singleton application instance
                theApp = cls.instance()

                # Instantiate an original QApplication instance if we need to
                if theApp is None:
                    theApp = QtWidgets.real_QApplication(*args, **kwargs)
                    QtWidgets.qApp = theApp

                for key in ["__init__", "exec_", "quit"]:
                    if not hasattr(cls, key):
                        continue
                    # Skip attributes that we already have
                    val = getattr(cls, key)
                    if hasattr(theApp.__class__, key):
                        if hash(val) == hash(getattr(theApp.__class__, key)):
                            continue
                    # Make method?
                    if callable(val):
                        if hasattr(val, "im_func"):
                            val = val.im_func  # Python 2.x
                        val = types.MethodType(val, theApp.__class__)
                    # Set attribute on app instance (not the class!)
                    try:
                        setattr(theApp, key, val)
                    except Exception:
                        pass  # tough luck

                # "exec" as an attribute would cause a syntax error in Python 2.7
                # because it is a keyword.
                # So we have to use setattr instead of assigning theApp.exec
                setattr(theApp, "exec", theApp.exec_)  # noqa: B010

                # Call init function (in case the user overloaded it)
                theApp.__init__(*args, **kwargs)

                # Return global app object (modified to the users needs)
                return theApp

            def __init__(self, *args, **kwargs):
                pass

            def exec_(self, *args, **kwargs):
                """This function does nothing, except printing a
                warning message. The point is that a Qt App can crash
                quite hard if an object goes out of scope, and the error
                is not obvious.
                """
                print_mainloop_warning(mainloopWarning_qt)

                # Store local namespaces (scopes) of any functions that
                # precede this call. It might have a widget or application
                # object that should not be deleted ...
                import inspect
                import __main__

                for caller in inspect.stack()[1:]:
                    frame, name = caller[0], caller[3]
                    if name.startswith("<"):  # most probably "<module>"
                        break
                    else:
                        __main__.__dict__[name + "_locals"] = frame.f_locals

                # Tell interpreter to ignore any system exits
                sys._pyzoInterpreter.ignore_sys_exit = True

                # But re-enable it as soon as *this event* is processed
                def reEnableSysExit():
                    sys._pyzoInterpreter.ignore_sys_exit = False

                self._reEnableSysExitTimer = timer = QtCore.QTimer()
                timer.singleShot(0, reEnableSysExit)

            def quit(self, *args, **kwargs):
                """Do not quit if Qt app quits."""
                pass

        # Instantiate application object
        self.app = QApplication_hijacked([""])

        # Keep it alive even if all windows are closed
        self.app.setQuitOnLastWindowClosed(False)

        # Replace app class
        QtWidgets.QApplication = QApplication_hijacked

        # Notify that we integrated the event loop
        self.app._in_event_loop = "Pyzo"
        QtWidgets._in_event_loop = "Pyzo"

        # Use sys.excepthook to catch keyboard interrupts that occur
        # in event handlers. We also want to call the current hook
        self._original_excepthook = sys.excepthook
        sys.excepthook = self._excepthook

    def _excepthook(self, type, value, traceback):
        if issubclass(type, KeyboardInterrupt):
            self._keyboard_interrupt()
        elif self._original_excepthook is not None:
            return self._original_excepthook(type, value, traceback)

    def process_events(self):
        self.app.sendPostedEvents()
        self.app.processEvents()

    def _workaroundWindowsInBackground(self):
        """prevent windows/dialogs from spawning in the background

        In MS Windows 10/11, Qt windows most times do not appear at the front.
        For example the command "QtWidgets.QMessageBox().exec()" will create a
        message box window, but this will not be active and might even be covered
        by the Pyzo GUI's window.
        On the one hand, MS Windows does this to prevent focus stealing.
        On the other hand, people get confused when they do not see a plot window
        appear on the screen after a plot command. Such Qt windows are grouped
        together with the Pyzo IDE in the operating system's taskbar, making it
        more difficult to see that there is a dialog waiting for user input.

        As a workaround, we create a first dummy window that we show minimized,
        then in normal state before closing it again. All subsequent commands to
        display a dialog, window, etc. will then create the window in the foreground
        because of this initial focus.
        """
        if sys.platform == "win32":
            QtCore, QtWidgets = self._QtCore, self._QtWidgets
            w = QtWidgets.QWidget()
            w.setWindowFlags(QtCore.Qt.WindowType.FramelessWindowHint)
            w.showMinimized()
            w.setWindowState(QtCore.Qt.WindowState.WindowActive)
            w.close()

    def run(self, repl_callback):
        # Create timer
        timer = self._timer = self._QtCore.QTimer()
        timer.setSingleShot(False)
        timer.setInterval(int(0.1 * 1000))  # ms
        timer.timeout.connect(repl_callback)
        timer.start()

        self._workaroundWindowsInBackground()

        # Enter Qt mainloop

        # Qt5 still has "exec_" because of the "exec" keyword in Python 2
        exec_ = getattr(self._QtWidgets.real_QApplication, "exec", None)
        if exec_ is None:
            exec_ = self._QtWidgets.real_QApplication.exec_
        exec_()

    def quit(self):
        # A nicer way to quit
        self._QtWidgets.real_QApplication.quit()


class App_pyqt6(App_qt):
    """Hijack the PyQt6 mainloop."""

    _API = "PyQt6"


class App_pyqt5(App_qt):
    """Hijack the PyQt5 mainloop."""

    _API = "PyQt5"


class App_pyside6(App_qt):
    """Hijack the PySide6 mainloop."""

    _API = "PySide6"


class App_pyside2(App_qt):
    """Hijack the PySide2 mainloop."""

    _API = "PySide2"


class App_wx(App_base):
    """Hijack the wxWidgets mainloop."""

    def __init__(self):
        # Try importing
        try:
            import wx
        except ImportError:
            # For very old versions of WX
            import wxPython as wx

        # Create dummy mainloop to replace original mainloop
        def dummy_mainloop(*args, **kw):
            print_mainloop_warning()

        # Depending on version, replace mainloop
        ver = wx.__version__
        orig_mainloop = None
        if ver[:3] >= "2.5":
            if hasattr(wx, "_core_"):
                core = wx._core_
            elif hasattr(wx, "_core"):
                core = wx._core
            else:
                raise ImportError
            orig_mainloop = core.PyApp_MainLoop
            core.PyApp_MainLoop = dummy_mainloop
        elif ver[:3] == "2.4":
            orig_mainloop = wx.wxc.wxPyApp_MainLoop
            wx.wxc.wxPyApp_MainLoop = dummy_mainloop
        else:
            # Unable to find either wxPython version 2.4 or >= 2.5."
            raise ImportError

        self._orig_mainloop = orig_mainloop
        # Store package wx
        self.wx = wx

        # Get and store the app instance to process events
        app = wx.GetApp()
        if app is None:
            app = wx.App(False)
        self.app = app

        # Notify that we integrated the event loop
        self.app._in_event_loop = "Pyzo"
        wx._in_event_loop = "Pyzo"

    def process_events(self):
        wx = self.wx

        # This bit is really needed
        old = wx.EventLoop.GetActive()
        eventLoop = wx.EventLoop()
        wx.EventLoop.SetActive(eventLoop)
        while eventLoop.Pending():
            eventLoop.Dispatch()

        # Process and reset
        self.app.ProcessIdle()  # otherwise frames do not close
        wx.EventLoop.SetActive(old)


class App_gtk(App_base):
    """Modifies pyGTK's mainloop with a dummy so user code does not
    block IPython.  processing events is done using the module'
    main_iteration function.
    """

    def __init__(self):
        # Try importing gtk
        import gtk

        # Replace mainloop with a dummy
        def dummy_mainloop(*args, **kwargs):
            print_mainloop_warning()

        gtk.mainloop = dummy_mainloop
        gtk.main = dummy_mainloop

        # Replace main_quit with a dummy too
        def dummy_quit(*args, **kwargs):
            pass

        gtk.main_quit = dummy_quit
        gtk.mainquit = dummy_quit

        # Make sure main_iteration exists even on older versions
        if not hasattr(gtk, "main_iteration"):
            gtk.main_iteration = gtk.mainiteration

        # Store 'app object'
        self.app = gtk

        # Notify that we integrated the event loop
        self.app._in_event_loop = "Pyzo"

    def process_events(self):
        gtk = self.app
        while gtk.events_pending():
            gtk.main_iteration(False)



"""
Views

When starting Pyzo, the window and panel geometries, and the loaded tool panels
are automatically restored. And when closing Pyzo, these geometry and state
values are automatically stored to the configuration file. This view is defined
as the "[NORMAL]" view.

The ViewManager class makes it possible to have more than a single view setup
and to quickly switch between these defined views.

To switch between views, the "Views" dialog must be started, either via Pyzo's
menu, or by pressing the assigned key combination.

When invoked via a key combination that contains a modifier key (e.g. Ctrl),
the dialog will only remain open as long as at least one of these modifier keys
is kept pressed. A repeated press of the non-modifier key (e.g. the 'A' in
Ctrl+Shift+A) will advance the selection to the next entry in the dialog's list
widget, similar to the Pyzo editor's tab switching via Ctrl+Tab.
When releasing all modifier keys, the dialog will be closed and the selected
view in the dialog's list will be activated.

When invoked via the menu or a key shortcut without modifier key, the dialog
will stay open like a normal dialog window.

The list widget contains all defined views, ordered by how recently they were
used, with the currently active view on top. This makes it easier to switch
between two recently used views via the key combinations.

The state of the "[NORMAL]" view is always saved automatically before switching
to another view. All other views must be manually updated by pressing the
"Update view" button in the dialog (with the correct list entry selected).
"""

import sys
import base64

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa


ismacos = sys.platform.startswith("darwin")


class ViewManager:
    NORMAL_VIEW_NAME = "[NORMAL]"

    def __init__(self):
        self._currentViewName = self.NORMAL_VIEW_NAME

        # create entry for normal view, if not present
        self._getViewData(self.NORMAL_VIEW_NAME, True)

        self._bringViewNameToTopOfHistory()

    def getViewNameHistory(self):
        """returns the view names, with recently used views first"""
        return [v["name"] for v in pyzo.config.state.views]

    def _getViewData(self, viewName, createIfNotFound=False):
        for viewData in pyzo.config.state.views:
            if viewData["name"] == viewName:
                break
        else:
            if createIfNotFound:
                viewData = {"name": viewName}
                pyzo.config.state.views.append(viewData)
            else:
                viewData = None

        if viewData:
            if viewName == self.NORMAL_VIEW_NAME:
                viewData = {
                    "name": viewName,
                    "loadedTools": pyzo.config.state.loadedTools,
                    "windowGeometry": pyzo.config.state.windowGeometry,
                    "windowState": pyzo.config.state.windowState,
                }
            return viewData

        raise KeyError("view not found: {}".format(repr(viewName)))

    def getCurrentViewName(self):
        return self._currentViewName

    def isNormalViewActive(self):
        return self._currentViewName == self.NORMAL_VIEW_NAME

    def _bringViewNameToTopOfHistory(self):
        """move current view to front of view list, and keep order of other elements"""
        viewName = self._currentViewName
        pyzo.config.state.views.sort(key=lambda v: v["name"] != viewName)

    def selectView(self, viewName):
        if self.isNormalViewActive() and viewName != self.NORMAL_VIEW_NAME:
            pyzo.main.saveWindowState()

        # Restoring views in Qt is tricky. I had to come up with the following
        # workarounds to make it work with all Qt5 and Qt6 wrappers.
        # There is no guarantee that this will always work. If the geometries are
        # not properly restored, try again by switching to another view and back.
        # The following code worked fine all the time during my tests.
        # Altered versions of the code might also work but with a lower success
        # rate, so do many tests with the new version after "optimizing" the code.

        if self._currentViewName != viewName:
            viewData = self._getViewData(viewName)

            pyzo.main.restoreTools(viewData["loadedTools"])

            def doLater():
                pyzo.main.restoreGeometry(base64.b64decode(viewData["windowGeometry"]))
                pyzo.main.restoreState(base64.b64decode(viewData["windowState"]))
                pyzo.main.paintNow()
                pyzo.main.restoreGeometry(base64.b64decode(viewData["windowGeometry"]))
                pyzo.main.restoreState(base64.b64decode(viewData["windowState"]))

            self._timer = QtCore.QTimer()
            self._timer.singleShot(200, doLater)

        self._currentViewName = viewName
        self._bringViewNameToTopOfHistory()

    def saveView(self, viewName):
        if viewName == self.NORMAL_VIEW_NAME:
            pyzo.main.saveWindowState()
        else:
            viewData = self._getViewData(viewName, createIfNotFound=True)
            viewData.update(pyzo.main.getWindowState())

    def removeView(self, viewName):
        if viewName == self.NORMAL_VIEW_NAME:
            return False

        for v in pyzo.config.state.views[:]:
            if v["name"] == viewName:
                pyzo.config.state.views.remove(v)
        if viewName == self._currentViewName:
            self.selectView(self.NORMAL_VIEW_NAME)

        return True

    def renameView(self, oldName, newName):
        if oldName == self.NORMAL_VIEW_NAME:
            return False
        if newName in self.getViewNameHistory():
            return False
        for v in pyzo.config.state.views:
            if v["name"] == oldName:
                v["name"] = newName
                if oldName == self._currentViewName:
                    self._currentViewName = newName
                break


class QListWidgetModified(QtWidgets.QListWidget):
    def changeSelection(self, delta):
        row = self.currentRow()
        if row == -1:
            row = 0
        self.setCurrentRow((row + delta) % self.count())

    def wheelEvent(self, event):
        # change the list selection by moving the mouse wheel
        degrees = event.angleDelta().y() / 8.0
        delta = -1 if degrees > 0 else (1 if degrees < 0 else 0)
        if delta == 0:
            return
        self.changeSelection(delta)


class ViewManagerDialog(QtWidgets.QDialog):
    @classmethod
    def registerMenuAction(cls, action):
        cls._menuAction = action

    def __init__(self, parent):
        super().__init__(parent)

        self._finished = False
        self._modifiersWerePressedWhenStarted = None

        self.setWindowTitle("Views")

        layout = QtWidgets.QHBoxLayout()
        self.setLayout(layout)
        layoutLeft = QtWidgets.QVBoxLayout()
        layoutRight = QtWidgets.QVBoxLayout()
        layout.addLayout(layoutLeft, 1)
        layout.addLayout(layoutRight, 1)

        # right layout column
        layout = layoutRight

        w = self._btnUpdateView = QtWidgets.QPushButton("Update view")
        w.clicked.connect(self._btnUpdateViewClicked)
        w.setToolTip(
            "Save the current view to the selected view.\n"
            'The "[NORMAL]" view is automatically updated before activating another view.'
        )
        layout.addWidget(w)

        w = self._btnRenameView = QtWidgets.QPushButton("Rename view")
        w.clicked.connect(self._btnRenameViewClicked)
        w.setToolTip(
            'Rename the selected view.\nThe "[NORMAL]" view cannot be renamed.'
        )
        layout.addWidget(w)

        w = self._btnRemoveView = QtWidgets.QPushButton("Remove view")
        w.clicked.connect(self._btnRemoveViewClicked)
        w.setToolTip(
            'Remove the selected view.\nThe "[NORMAL]" view cannot be removed.'
        )
        layout.addWidget(w)

        w = self._btnAddView = QtWidgets.QPushButton("Add view")
        w.clicked.connect(self._btnAddViewClicked)
        w.setToolTip("Add a new view and activate it.")
        layout.addWidget(w)

        w = self._btnRenameView = QtWidgets.QPushButton("Activate view")
        w.clicked.connect(self._btnActivateViewClicked)
        w.setToolTip("Activate the selected view.")
        layout.addWidget(w)

        w = self._btnRenameView = QtWidgets.QPushButton("Cancel")
        w.clicked.connect(self.close)
        layout.addWidget(w)

        layout.addStretch()

        # left layout column
        layout = layoutLeft

        lw = self._lw = QListWidgetModified()
        layout.addWidget(lw)

        # set up the list widget
        self._entries = pyzo.viewManager.getViewNameHistory()

        lw = self._lw
        lw.clear()
        for s in self._entries:
            label = s
            lw.addItem(QtWidgets.QListWidgetItem(label))

        lw.setCurrentRow(0)
        lw.setMinimumWidth(
            lw.sizeHintForColumn(0) + lw.verticalScrollBar().sizeHint().width() + 10
        )

        lw.setToolTip(
            "list of available views, ordered by how recently they were used\n"
            "with the currently active view on top"
        )
        lw.itemDoubleClicked.connect(self._listItemDoubleClicked)

    def _inputNewName(self, initialText):
        text, ok = QtWidgets.QInputDialog.getText(
            self,
            "Pyzo views",
            "Enter new view name:",
            QtWidgets.QLineEdit.EchoMode.Normal,
            initialText,
        )
        return text, ok

    def _makeUniqueName(self, name):
        hist = pyzo.viewManager.getViewNameHistory()
        while name in hist:
            name += " (2)"
        return name

    def _btnRemoveViewClicked(self):
        viewName = self._lw.currentItem().text()
        if viewName == ViewManager.NORMAL_VIEW_NAME:
            return
        self.close()
        self._finished = True
        pyzo.viewManager.removeView(viewName)

    def _btnRenameViewClicked(self):
        oldName = self._lw.currentItem().text()
        if oldName == ViewManager.NORMAL_VIEW_NAME:
            return
        newName, ok = self._inputNewName(oldName)
        if not ok:
            return
        self.close()
        self._finished = True
        newName = self._makeUniqueName(newName)
        pyzo.viewManager.renameView(oldName, newName)

    def _btnAddViewClicked(self):
        newName, ok = self._inputNewName("")
        if not ok:
            return
        newName = self._makeUniqueName(newName)
        self.close()
        self._finished = True
        pyzo.viewManager.saveView(newName)
        pyzo.viewManager.selectView(newName)

    def _btnUpdateViewClicked(self):
        viewName = self._lw.currentItem().text()
        self.close()
        self._finished = True
        pyzo.viewManager.saveView(viewName)

    def _btnActivateViewClicked(self):
        self._finishSelection()

    def _listItemDoubleClicked(self, item):
        self._finishSelection()

    def _finishSelection(self):
        viewName = self._lw.currentItem().text()
        self.close()
        self._finished = True
        pyzo.viewManager.selectView(viewName)

    def keyPressEvent(self, event):
        if self._modifiersWerePressedWhenStarted is None:
            self._modifiersWerePressedWhenStarted = False

        shortcuts = self._menuAction.shortcut()
        if self._finished:
            return

        if pyzo.qt.QT_VERSION_STR.startswith("5."):
            kc = event.key() + int(event.modifiers())  # int
        else:
            kc = event.keyCombination()  # QKeyCombination
        for sc in shortcuts:
            if kc == sc:
                self._lw.changeSelection(1)
                return  # consume event
        super().keyPressEvent(event)

    def keyReleaseEvent(self, event):
        if self._modifiersWerePressedWhenStarted is None:
            if event.modifiers():
                self._modifiersWerePressedWhenStarted = True

        if self._modifiersWerePressedWhenStarted:
            if not event.modifiers():
                self._finishSelection()
                return  # consume event

        super().keyReleaseEvent(event)



import sys
import re
import html
from functools import partial

from pyzo.qt import QtCore, QtGui, QtWidgets
import pyzo

tool_name = pyzo.translate("pyzoInteractiveHelp", "Interactive help")
tool_summary = pyzo.translate(
    "pyzoInteractiveHelp", "Shows help on an object when using up/down in autocomplete."
)

keywordsHelp = {
    "await": pyzo.translate(
        "pyzoInteractiveHelp",
        "Suspend the execution of coroutine on an awaitable object. Can only be used inside a coroutine function.",
    ),
    "else": pyzo.translate(
        "pyzoInteractiveHelp",
        """This keyword can appear as part of an alternative (see: ``if``), a loop (see: ``for``, ``while``) or a ``try`` statement.""",
    ),
    "import": pyzo.translate(
        "pyzoInteractiveHelp",
        """Find and load modules or members of modules.

Examples:
```
import foo                 # foo imported and bound locally
import foo.bar.baz         # foo.bar.baz imported, foo bound locally
import foo.bar.baz as fbb  # foo.bar.baz imported and bound as fbb
from foo.bar import baz    # foo.bar.baz imported and bound as baz
from foo import attr       # foo imported and foo.attr bound as attr
from foo import *          # foo imported and all its members bound
                                    # under their respective names
```""",
    ),
    "pass": pyzo.translate(
        "pyzoInteractiveHelp",
        "This is an instruction that does nothing. It is useful in cases where a statement must appear because of syntactic constraints but we have nothing to do.",
    ),
    "break": pyzo.translate(
        "pyzoInteractiveHelp",
        """This keyword can only appear in the body of a ``for`` or ``while`` loop. It terminates the loop early, regardless of the loop condition.

See also: ``continue``""",
    ),
    "except": pyzo.translate(
        "pyzoInteractiveHelp",
        "An ``except`` clause can appear as a part of a ``try`` statement.",
    ),
    "in": pyzo.translate(
        "pyzoInteractiveHelp",
        """This keyword usually refers to membership of an object in a structure. There are actually two different kinds of ``in``.

In a construct of the form ``for identifier in iterable``, ``in`` is a purely syntactic element that bears no meaning per se. See: ``for``

Outside such constructs, ``in`` is an operator and its precise meaning depends on the type of the first operand.""",
    ),
    "raise": pyzo.translate(
        "pyzoInteractiveHelp",
        "The ``raise`` statement is used to raise an Exception. Raising an exception will cause the program to abort, unless the exception is handled by a surrounding ``try``/``except`` statement or ``with`` statement.",
    ),
    "class": pyzo.translate(
        "pyzoInteractiveHelp", "This keyword introduces the definition of a class."
    ),
    "finally": pyzo.translate(
        "pyzoInteractiveHelp",
        "An ``finally`` clause can appear as a part of a ``try`` statement.",
    ),
    "is": pyzo.translate(
        "pyzoInteractiveHelp",
        """This operator tests for an object's identity: ``x is y`` is true if and only if ``x`` and ``y`` are the same object.

See also: ``id``""",
    ),
    "return": pyzo.translate(
        "pyzoInteractiveHelp",
        "This keyword can only appear in the definition of a function. It is usually followed by an expression. It means that the execution of the function call terminates here and that the result of the call is the value of the following expression. If ``return`` is not followed by an expression, the result is ``None``.",
    ),
    "and": pyzo.translate(
        "pyzoInteractiveHelp",
        """This operator computes the boolean conjunction in a lazy manner. More precisely, the expression ``x and y`` first evaluates ``x``; if ``x`` is false, its value is returned; otherwise, ``y`` is evaluated and the resulting value is returned.

See also: ``or``, ``not``, ``True``, ``False``""",
    ),
    "continue": pyzo.translate(
        "pyzoInteractiveHelp",
        """This keyword can only appear in the body of a ``for`` or ``while`` loop. It terminates the current run of the body early. The loop may still make additional runs of its body if its condition is still true .

See also: ``break``""",
    ),
    "for": pyzo.translate(
        "pyzoInteractiveHelp",
        """The ``for`` statement is used to iterate over the elements of an iterable object:

```
for variable in iterable:
    body
else:       # optional
    suite
```
The expression ``iterable`` is evaluated once; its value should be an iterable object. An iterator is created for the result of that expression. The ``body`` is then executed once for each item provided by the iterator, in the order returned by the iterator: each item in turn is assigned to the ``variable`` and then the ``body`` is executed. When the items are exhausted (which is immediately when the ``iterable`` is empty), the ``suite`` in the ``else`` clause, if present, is executed, and the loop terminates.

A ``break`` statement executed in the first suite terminates the loop without executing the ``else`` clause's ``suite``. A ``continue`` statement executed in the first suite skips the rest of the ``body`` and continues with the next item, or with the ``else`` clause if there is no next item.

The for-loop makes assignments to the ``variable``. This overwrites all previous assignments to that variable including those made in the ``body`` of the for-loop:
```
for i in range(10):
    print(i)
    i = 5             # this will not affect the for-loop
                      # because i will be overwritten with the next
                      # index in the range
```

The ``for`` keyword can also be part of a generator.

Example:
```
[2*x for x in T]
```
This builds from ``T`` a list where each item was doubled.
""",
    ),
    "lambda": pyzo.translate(
        "pyzoInteractiveHelp",
        """This keyword is used to produce an anonymous function.

Example:
```
lambda x,y,z : (x+y) * z
```""",
    ),
    "try": pyzo.translate(
        "pyzoInteractiveHelp",
        """The ``try`` statement allows to recover when an Exception was raised. It has two main forms.

```
try:
    critical_section
except SomeExceptionType as e:
    suite   # manage exceptions of type SomeExceptionType
else:       # optional
    everything_ok
```
If ``critical_section`` raises an exception of type ``SomeExceptionType``, the whole program will not abort but the ``suite`` will be executed, with ``e`` being bound to the actual exception. Several ``except`` clauses may be specified, each dealing with some kind of exception. If an exception is raised in ``critical_section`` but is not managed by any ``except`` clause, the whole program aborts. If there is an ``else`` clause, its body is executed if the control flow leaves the ``critical_section``, no exception was raised, and no ``return``, ``continue`` or ``break`` statement was executed. Exceptions in the ``else`` clause are not handled by the preceding ``except`` clauses.


```
try:
    critical_section
finally:
    suite
```
If ``critical_section`` raises an exception, the ``suite`` will be executed before aborting the program. The ``suite`` is also executed when no exception is raised. When a ``return``, ``break`` or ``continue`` statement is executed in the ``critical_section``, the ``finally`` clause is also executed 'on the way out.' A ``continue`` statement is illegal in the ``finally`` clause.

Both mechanisms can be combined, the ``finally`` clause coming after the last ``except`` clause or after the optional ``else`` clause.""",
    ),
    "as": pyzo.translate(
        "pyzoInteractiveHelp",
        "The ``as`` keyword can appear in an ``import`` statement, in an ``except`` clause of a ``try`` statement or in a ``with`` statement.",
    ),
    "def": pyzo.translate(
        "pyzoInteractiveHelp", "This keyword introduces the definition of a function."
    ),
    "from": pyzo.translate(
        "pyzoInteractiveHelp",
        "This keyword can appear as a part of an ``import`` or a ``raise``. It has different meanings; see help on those keywords.",
    ),
    "nonlocal": pyzo.translate(
        "pyzoInteractiveHelp",
        """This keyword can only appear in the definition of a function. It is followed by identifiers and indicates that those identifier refer variables from the outer scope, not local variables in the function being defined.

See also: ``global``""",
    ),
    "while": pyzo.translate(
        "pyzoInteractiveHelp",
        """The while statement is used for repeated execution as long as an expression is true:

```
while expression:
    body
else:           # optional
    suite
```
This repeatedly tests the ``expression`` and, if it is true, executes the ``body``; if the ``expression`` is false (which may be the first time it is tested) the ``suite`` of the ``else`` clause, if present, is executed and the loop terminates.

A ``break`` statement executed in the ``body`` terminates the loop without executing the ``else`` clause's ``suite``. A ``continue`` statement executed in the ``body`` skips the rest of the ``body`` and goes back to testing the ``expression``.""",
    ),
    "assert": pyzo.translate(
        "pyzoInteractiveHelp",
        "Assert statements are a convenient way to insert debugging assertions into a program.",
    ),
    "del": pyzo.translate(
        "pyzoInteractiveHelp",
        """Deletion of a name removes the binding of that name from the local or global namespace. It is also possible to delete an item in a list.""",
    ),
    "global": pyzo.translate(
        "pyzoInteractiveHelp",
        """This keyword is followed by identifiers and indicates that those identifiers refer variables from the module scope, not local variables.

See also: ``nonlocal``""",
    ),
    "not": pyzo.translate(
        "pyzoInteractiveHelp",
        """The operator ``not`` returns ``True`` if its argument is false, ``False`` otherwise.

See also: ``and``, ``or``, ``True``, ``False``""",
    ),
    "with": pyzo.translate(
        "pyzoInteractiveHelp",
        """The ``with`` statement is used to wrap the execution of a block with methods defined by a context manager.

Example:
```
with open(filename) as infile:
    header = infile.readline()
    # etc.
```
In this example, the context manager will ensure correct closing of the file upon exiting the ``with`` statement, even if an exception was raised.""",
    ),
    "async": pyzo.translate(
        "pyzoInteractiveHelp",
        "This keyword appears as part of the ``async def``, ``async for`` and ``async with`` constructs and allows for writing coroutines.",
    ),
    "elif": pyzo.translate(
        "pyzoInteractiveHelp",
        "This keyword can only appear as part of an alternative. See: ``if``",
    ),
    "if": pyzo.translate(
        "pyzoInteractiveHelp",
        """This keyword usually introduces an alternative. It is followed by a conditon and introduces the statement to execute when this condition is true. The alternative can only have one ``if`` clause, at the beginning. It can also comprise one or more ``elif`` clauses and at most one ``else`` clause at the end.

Example:

```
if condition1 :
    doSomething       # when condition1 evaluates to True
elif condition2 :     # optional, may have several such clauses
    doSomeOtherthing  # when condition1 evaluates to False and condition2 to True
else :                # optional
    doAnotherThing    # when all conditions evaluate to False
```


The ``if`` keyword can also appear in a generator.

Example:
```
[x for x in T if x%2==0]
```
This filters ``T`` to keep only its even items.""",
    ),
    "or": pyzo.translate(
        "pyzoInteractiveHelp",
        """This operator computes the boolean disjunction in a lazy manner. More precisely, the expression ``x or y`` first evaluates ``x``; if ``x`` is true, its value is returned; otherwise, ``y`` is evaluated and the resulting value is returned.

See also: ``and``, ``not``, ``True``, ``False``""",
    ),
    "yield": pyzo.translate(
        "pyzoInteractiveHelp",
        """The ``yield`` expression is used when defining a generator function or an asynchronous generator function and thus can only be used in the body of a function definition. Using a yield expression in a function's body causes that function to be a generator, and using it in an ``async def`` function's body causes that coroutine function to be an asynchronous generator.""",
    ),
}

operators = [
    "+",
    "-",
    "*",
    "**",
    "/",
    "//",
    "%",
    "@",
    "<<",
    ">>",
    "&",
    "|",
    "^",
    "~",
    "<",
    ">",
    "<=",
    ">=",
    "==",
    "!=",
]
operatorsHelp = pyzo.translate(
    "pyzoInteractiveHelp",
    "No help is available for operators because they are ambiguous: their meaning depend on the type of the first operand.",
)


#
htmlWrap = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN"
"http://www.w3.org/TR/REC-html40/strict.dtd">
<html>
<head>
<style type="text/css">
pre, code {{background-color: BGCOLOR_CODE;}}
</style>
</head>
<body style=" font-family:'Sans Serif'; font-size:{}pt; font-weight:400; font-style:normal;">
{}
</body>
</html>
"""


# Define title text (font-size percentage does not seem to work sadly.)
def get_title_text(objectName, h_class="", h_repr=""):
    title_bgcolor = "#468" if pyzo.darkQt else "#def"
    title_text = "<p style='background-color:{};'>".format(title_bgcolor)
    if h_class == "~python_keyword~":
        title_text += "<b>Keyword:</b> {}".format(objectName)
    elif h_class == "~python_operator~":
        title_text += "<b>Operator:</b> {}".format(objectName)
    elif h_class == "":
        title_text += "<b>Unknown construct:</b> {}".format(objectName)
    else:
        title_text += "<b>Object:</b> {}".format(objectName)
        title_text += ", <b>class:</b> {}".format(h_class)
        if h_repr:
            if len(h_repr) > 40:
                h_repr = h_repr[:37] + "..."
            title_text += ", <b>repr:</b> {}".format(h_repr)

    # Finish
    title_text += "</p>\n"
    return title_text


initText = pyzo.translate(
    "pyzoInteractiveHelp",
    """
Help information is queried from the current shell
automatically when moving up/down in the autocompletion list,
or manually via the context menu in the shell and editor,
or via the Workspace tool.
""",
)


class PyzoInteractiveHelpHistoryMenu(QtWidgets.QMenu):
    def __init__(self, title, parent, forward):
        super().__init__(title, parent)
        self._forward = forward
        self.aboutToShow.connect(self.populate)

    def populate(self):
        self.clear()
        if self._forward:
            indices = range(self.parent()._histindex + 1, len(self.parent()._history))
        else:
            indices = range(self.parent()._histindex - 1, -1, -1)
        for i in indices:
            action = self.addAction(self.parent()._history[i])
            action.triggered.connect(partial(self.doAction, i=i))

    def doAction(self, i):
        self.parent()._histindex = i
        self.parent().setObjectName(self.parent().currentHist())


class PyzoInteractiveHelp(QtWidgets.QWidget):
    def __init__(self, parent):
        super().__init__(parent)

        # Create text field, checkbox, and button
        self._text = QtWidgets.QLineEdit(self)
        self._printBut = QtWidgets.QPushButton("Print", self)

        style = QtWidgets.qApp.style()

        self._backBut = QtWidgets.QToolButton(self)
        self._backBut.setIcon(style.standardIcon(style.StandardPixmap.SP_ArrowLeft))
        self._backBut.setIconSize(QtCore.QSize(16, 16))
        self._backBut.setPopupMode(self._backBut.ToolButtonPopupMode.DelayedPopup)
        self._backBut.setMenu(
            PyzoInteractiveHelpHistoryMenu("Backward menu", self, False)
        )

        self._forwBut = QtWidgets.QToolButton(self)
        self._forwBut.setIcon(style.standardIcon(style.StandardPixmap.SP_ArrowRight))
        self._forwBut.setIconSize(QtCore.QSize(16, 16))
        self._forwBut.setPopupMode(self._forwBut.ToolButtonPopupMode.DelayedPopup)
        self._forwBut.setMenu(
            PyzoInteractiveHelpHistoryMenu("Forward menu", self, True)
        )

        self._pauseBut = QtWidgets.QToolButton(self)
        self._pauseBut.setIcon(style.standardIcon(style.StandardPixmap.SP_MediaPause))
        self._pauseBut.setIconSize(QtCore.QSize(16, 16))
        self._pauseBut.setCheckable(True)
        self._pauseBut.setToolTip("Pause automatic help from autocompletion")

        # Create options button
        self._options = QtWidgets.QToolButton(self)
        self._options.setIcon(pyzo.icons.wrench)
        self._options.setIconSize(QtCore.QSize(16, 16))
        self._options.setPopupMode(self._options.ToolButtonPopupMode.InstantPopup)
        self._options.setToolButtonStyle(
            QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon
        )

        # Create options menu
        self._options._menu = QtWidgets.QMenu()
        self._options.setMenu(self._options._menu)

        # Create browser
        self._browser = QtWidgets.QTextBrowser(self)
        self._browser_text = initText
        self._browser.setContextMenuPolicy(
            QtCore.Qt.ContextMenuPolicy.CustomContextMenu
        )
        self._browser.customContextMenuRequested.connect(self.showMenu)

        # Create two sizers
        self._sizer1 = QtWidgets.QVBoxLayout(self)
        self._sizer2 = QtWidgets.QHBoxLayout()

        # Put the elements together
        self._sizer2.addWidget(self._backBut, 1)
        self._sizer2.addWidget(self._forwBut, 2)
        self._sizer2.addWidget(self._pauseBut, 1)
        self._sizer2.addWidget(self._text, 4)
        self._sizer2.addWidget(self._printBut, 0)
        self._sizer2.addWidget(self._options, 3)
        #
        self._sizer1.addLayout(self._sizer2, 0)
        self._sizer1.addWidget(self._browser, 1)
        #
        self._sizer1.setSpacing(2)
        # set margins
        margin = pyzo.config.view.widgetMargin
        self._sizer1.setContentsMargins(margin, margin, margin, margin)

        self.setLayout(self._sizer1)

        # Set config
        toolId = self.__class__.__name__.lower()
        self._config = config = pyzo.config.tools[toolId]
        #
        if not hasattr(config, "smartNewlines"):
            config.smartNewlines = True
        if not hasattr(config, "fontSize"):
            if sys.platform == "darwin":
                config.fontSize = 12
            else:
                config.fontSize = 10

        # Create callbacks
        self._text.returnPressed.connect(self.queryDoc)
        self._printBut.clicked.connect(self.printDoc)
        self._backBut.clicked.connect(self.goBack)
        self._forwBut.clicked.connect(self.goForward)
        self._pauseBut.clicked.connect(self.onTogglePause)
        #
        self._options.pressed.connect(self.onOptionsPress)
        self._options._menu.triggered.connect(self.onOptionMenuTiggered)

        # Start
        self._history = []
        self._histindex = 0
        self.setText()  # Set default text
        self.onOptionsPress()  # Fill menu

        # Set temporary lookup texts from editor/shell extensions.
        # The order of the entries is also the priority (first found will be displayed).
        self._extensionLookups = {s: None for s in ["autocomplete", "calltip"]}

    def showMenu(self, pos):
        menu = self._browser.createStandardContextMenu()
        help = QtGui.QAction(
            pyzo.icons.help, pyzo.translate("pyzoInteractiveHelp", "Help on this"), menu
        )
        help.triggered.connect(partial(self.helpOnThis, pos=pos))
        menu.insertAction(menu.actions()[0], help)
        menu.exec(self.mapToGlobal(pos))

    def helpOnThis(self, pos):
        name = self._browser.textCursor().selectedText().strip()
        if name == "":
            cursor = self._browser.cursorForPosition(pos)
            cursor.select(cursor.SelectionType.WordUnderCursor)
            name = cursor.selectedText()
        if name != "":
            self.setObjectName(name, True)

    def onOptionsPress(self):
        """Create the menu for the button, Do each time to make sure
        the checks are right.
        """

        # Get menu
        menu = self._options._menu
        menu.clear()

        # Add smart format option
        action = menu.addAction(pyzo.translate("pyzoInteractiveHelp", "Smart format"))
        action._what = "smart"
        action.setCheckable(True)
        action.setChecked(bool(self._config.smartNewlines))

        # Add delimiter
        menu.addSeparator()

        # Add font size options
        currentSize = self._config.fontSize
        for i in range(8, 15):
            action = menu.addAction(
                pyzo.translate("pyzoInteractiveHelp", "Font size: {}").format(i)
            )
            action._what = "font-size: {}px".format(i)
            action.setCheckable(True)
            action.setChecked(i == currentSize)

    def onOptionMenuTiggered(self, action):
        """The user decides what to show in the structure."""

        # Get text
        text = action._what.lower()

        if "smart" in text:
            # Swap value
            current = bool(self._config.smartNewlines)
            self._config.smartNewlines = not current
            # Update
            self.queryDoc()

        elif "size" in text:
            # Get font size
            size = int(text.split(":", 1)[1][:-2])
            # Update
            self._config.fontSize = size
            # Update
            self.setText()

    def setText(self, text=None):
        # (Re)store text
        if text is None:
            text = self._browser_text
        else:
            self._browser_text = text

        # Set text with html header
        size = self._config.fontSize
        self._browser.setHtml(
            htmlWrap.replace(
                "BGCOLOR_CODE", "#505050" if pyzo.darkQt else "#f2f2f2"
            ).format(size, text)
        )

    def setObjectName(self, name, addToHist=False):
        """Set the object name programatically
        and query documentation for it.
        """
        self._text.setText(name)
        self.queryDoc(addToHist)

    def helpFromExtension(self, requestedName, source, addToHist=False):
        assert source in self._extensionLookups
        self._extensionLookups[source] = requestedName

        if self._pauseBut.isChecked():
            return

        for s in self._extensionLookups.values():
            if s is not None:
                name = s
                break
        else:
            name = self.currentHist()
            addToHist = False
        self.setObjectName(name, addToHist)

    def _escapeHtml(self, text):
        return html.escape(text).replace("\n", "<br />")

    def helpForCodeSnippet(self, snippet):
        if self._pauseBut.isChecked():
            return

        if snippet is None:
            self.setObjectName(self.currentHist() or "")
            return

        descriptionAsHtml = self._escapeHtml(snippet["description"])
        codeAsHtml = self._escapeHtml(snippet["code"])
        self._text.setText(snippet["name"])
        title_bgcolor = "#468" if pyzo.darkQt else "#def"
        text = (
            f"<p style='background-color:{title_bgcolor};'>"
            f"<b>Code snippet filepath:</b> {html.escape(snippet['filepath'])}"
            "</p>\n"
            f"{descriptionAsHtml}"
            f"<pre>{codeAsHtml}</pre>"
        )
        self.setText(text)

    def currentHist(self):
        try:
            return self._history[self._histindex]
        except Exception:
            return None

    def addToHist(self, name):
        if not name or name == self.currentHist():
            return
        self._history = self._history[: self._histindex + 1]
        self._history.append(name)
        self._histindex = len(self._history) - 1

    def goBack(self):
        if self._histindex > 0 and self._history != []:
            self._histindex -= 1
            self.setObjectName(self.currentHist())

    def goForward(self):
        if self._histindex < len(self._history) - 1:
            self._histindex += 1
            self.setObjectName(self.currentHist())

    def onTogglePause(self, pauseActivated):
        if pauseActivated:
            self.addToHist(self._text.text())

    def printDoc(self):
        """Print the doc for the text in the line edit."""
        # Get name
        name = self._text.text()
        # Tell shell to print doc
        shell = pyzo.shells.getCurrentShell()
        if shell and name:
            if name in operators:
                shell.processLine(
                    'print("""{}""")'.format(
                        "Help on operator: " + name + "\n\n" + operatorsHelp
                    )
                )
            elif name in keywordsHelp:
                shell.processLine(
                    'print("""{}""")'.format(
                        "Help on keyword: " + name + "\n\n" + keywordsHelp[name]
                    )
                )
            else:
                shell.processLine("print({}.__doc__)".format(name))

    def queryDoc(self, addToHistory=True):
        """Query the doc for the text in the line edit."""
        # Get name
        name = self._text.text()
        if addToHistory:
            self.addToHist(name)
        if name in operators:
            text = name + "\n~python_operator~\n\n\n" + operatorsHelp
            self.displayResponse(text)
        elif name in keywordsHelp:
            text = name + "\n~python_keyword~\n\n\n" + keywordsHelp[name]
            self.displayResponse(text)
        else:
            # Get shell and ask for the documentation
            shell = pyzo.shells.getCurrentShell()
            if shell and name:
                future = shell._request.doc(name)
                future.add_done_callback(self.queryDoc_response)
            elif not name:
                self.setText(initText)

    def queryDoc_response(self, future):
        """Process the response from the shell."""

        # Process future
        if future.cancelled():
            # print('Introspect cancelled') # No living kernel
            return
        elif future.exception():
            print("Introspect-queryDoc-exception: ", future.exception())
            return
        else:
            response = future.result()
            if not response:
                return
            self.displayResponse(response)

    def displayResponse(self, response):
        try:
            # Get parts
            parts = response.split("\n")
            objectName, h_class, h_fun, h_repr = tuple(parts[:4])
            h_text = "\n".join(parts[4:])

            # Obtain newlines that we hid for repr
            h_repr.replace("/r", "/n")

            # Make all newlines \n in h_text and strip
            h_text = h_text.replace("\r\n", "\n").replace("\r", "\n")
            h_text = h_text.lstrip()

            # Init text
            text = ""

            # These signs will fool the html
            h_repr = h_repr.replace("<", "&lt;")
            h_repr = h_repr.replace(">", "&gt;")
            h_text = h_text.replace("<", "&lt;")
            h_text = h_text.replace(">", "&gt;")

            if self._config.smartNewlines:
                # Make sure the signature is separated from the rest using at
                # least two newlines
                header = ""
                if True:
                    # Get short version of objectName
                    name = objectName.split(".")[-1]
                    # Is the signature in the docstring?
                    docs = h_text.replace("\n", "|")
                    tmp = re.search(
                        r"[a-zA-Z0-9_\.]*?" + re.escape(name) + r"\(.*?\)", docs
                    )
                    if tmp and tmp.span(0)[0] < 5:
                        header = tmp.group(0)
                        h_text = h_text[len(header) :].lstrip(":").lstrip()
                        header = header.replace("|", "")
                        # h_text = header + '\n\n' + h_text
                    elif h_text.startswith((objectName, name)):
                        header, sep, docs = h_text.partition("\n")
                        # h_text = header + '\n\n' + docs
                        h_text = docs
                    elif h_fun is not None and h_fun != "":
                        header = h_fun

                # Parse the text as rest/numpy like docstring
                h_text = self.smartFormat(h_text)
                h_text = re.sub(r"``(.*?)``", r"<code>\1</code>", h_text)
                if header:
                    h_text = "<p style='color:{};'><b>{}</b></p>\n{}".format(
                        "#0bf" if pyzo.darkQt else "#005",
                        header,
                        h_text,
                    )
                    # h_text = "<b>{}</b><br /><br />\n{}".format(header, h_text)
            else:
                # Make newlines html
                h_text = h_text.replace("\n", "<br />")

            # Compile rich text
            text += get_title_text(objectName, h_class, h_repr)
            if not self._config.smartNewlines and h_fun is not None and h_fun != "":
                text += "<p><b>Signature:</b> {}</p>".format(h_fun)
            text += "{}<br />".format(h_text)

        except Exception:
            try:
                text += get_title_text(objectName, h_class, h_repr)
                if h_fun is not None and h_fun != "":
                    text += "<p><b>Signature:</b> {}</p>".format(h_fun)
                text += h_text
            except Exception:
                text = response

        # Done
        # size = self._config.fontSize
        self.setText(text)

    def smartFormat(self, text):
        # Get lines
        lines = text.splitlines(True)

        # Test minimal indentation
        minIndent = 9999
        for line in lines[1:]:
            line_ = line.lstrip()
            indent = len(line) - len(line_)
            if line_:
                minIndent = min(minIndent, indent)

        # Remove minimal indentation
        lines2 = [lines[0]]
        for line in lines[1:]:
            lines2.append(line[minIndent:])

        # Prepare
        prevLine_ = ""
        prevIndent = 0
        prevWasHeader = False
        inExample = False
        forceNewline = False
        inPre = False

        # Format line by line
        lines3 = []
        for line in lines2:
            # Get indentation
            line_ = line.lstrip()
            if line_ in ("```", "```\n"):
                if inPre:
                    line = "</pre>"
                    inPre = False
                else:
                    line = "<pre>"
                    inPre = True
            elif not inPre:
                indent = len(line) - len(line_)
                # indentPart = line[:indent-minIndent]
                indentPart = line[:indent]

                if not line_:
                    lines3.append("<br />")
                    forceNewline = True
                    continue

                # Indent in html
                line = "&nbsp;" * len(indentPart) + line

                # Determine if we should introduce a newline
                isHeader = False
                if ("---" in line or "===" in line) and indent == prevIndent:
                    # Header
                    lines3[-1] = "<b>" + lines3[-1] + "</b>"
                    line = ""  #'<br /> ' + line
                    isHeader = True
                    inExample = False
                    # Special case, examples
                    if prevLine_.lower().startswith("example"):
                        inExample = True
                    else:
                        inExample = False
                elif " : " in line:
                    tmp = line.split(" : ", 1)
                    line = "<br /><u>" + tmp[0] + "</u> : " + tmp[1]
                elif line_.startswith("* "):
                    line = "<br />&nbsp;&nbsp;&nbsp;&#8226;" + line_[2:]
                elif prevWasHeader or inExample or forceNewline:
                    line = "<br />" + line
                else:
                    if prevLine_:
                        line = " " + line_
                    else:
                        line = line_

                # Force next line to be on a new line if using a colon
                if " : " in line:
                    forceNewline = True
                else:
                    forceNewline = False

                # Prepare for next line
                prevLine_ = line_
                prevIndent = indent
                prevWasHeader = isHeader

            # Done with line
            lines3.append(line)

        # Done formatting
        return "".join(lines3)



# Copyright (c) 2016, Almar Klein, Rob Reilink
#
# This file is distributed under the terms of the 2-Clause BSD License.

import os
import sys
from pyzo.qt import QtCore


def is_frozen():
    """Return whether this app is a frozen application (using e.g. cx_freeze)."""
    return bool(getattr(sys, "frozen", None))


def prepare_appdata_appconfig_dirs():
    """get the directories for Pyzo application data and configuration
    folders are created if not present
    """

    # check for a "settings" folder next to the Pyzo executable or one level above,
    # which is typically used for portable applications
    use_portable_settings = False
    if is_frozen():
        exec_dir = os.path.abspath(os.path.dirname(sys.executable))
        for reldir in ("_internal/settings", "settings", "../settings"):
            localpath = os.path.abspath(os.path.join(exec_dir, reldir))
            if os.path.isdir(localpath):
                try:
                    open(os.path.join(localpath, "test.write"), "wb").close()
                    os.remove(os.path.join(localpath, "test.write"))
                except IOError:
                    pass  # We cannot write in this directory
                else:
                    data_path = config_path = localpath
                    use_portable_settings = True
                    break

    if not use_portable_settings:
        appname = "pyzo"
        path_dot = os.path.expanduser(
            "~/." + appname
        )  # leading dot means hidden directory
        if sys.platform == "win32":
            data_path = config_path = os.path.join(os.getenv("APPDATA"), appname)
            # typically r"C:\Users\username\AppData\Roaming\pyzo"
        elif os.path.isdir(path_dot):
            # existing legacy data and config directory in Linux or macOS
            data_path = config_path = path_dot
        elif sys.platform == "linux":
            # see https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
            data_path_base = os.getenv("XDG_DATA_HOME", "") or "~/.local/share"
            config_path_base = os.getenv("XDG_CONFIG_HOME", "") or "~/.config"
            data_path = os.path.expanduser(os.path.join(data_path_base, appname))
            config_path = os.path.expanduser(os.path.join(config_path_base, appname))
        elif sys.platform == "darwin":
            sp = QtCore.QStandardPaths
            data_path_base = sp.writableLocation(sp.StandardLocation.AppDataLocation)
            config_path_base = sp.writableLocation(sp.StandardLocation.ConfigLocation)
            """
            The AppName added by Qt depends on how pyzo was started, e.g.:
                import pyzo; pyzo.start() --> ''
                python pyzolauncher.py --> 'pyzolauncher.py' or 'python'
                via binary from frozen pyzo --> 'pyzo'
            """
            if QtCore.QCoreApplication.applicationName() != "":
                data_path_base = os.path.split(data_path_base)[0]
                config_path_base = os.path.split(config_path_base)[0]
            data_path = os.path.join(data_path_base, appname)
            config_path = os.path.join(config_path_base, appname)
        else:
            raise NotImplementedError("unsupported operating system: " + sys.platform)

        os.makedirs(data_path, exist_ok=True)
        os.makedirs(config_path, exist_ok=True)

    # Create tooldir if necessary
    tool_dir = os.path.join(data_path, "tools")
    os.makedirs(tool_dir, exist_ok=True)

    return data_path, config_path



"""
PEP 768 introduced a new "safe external debugger interface" in
Python 3.14 -- see https://peps.python.org/pep-0768

This allows us to temporarily insert a Pyzo kernel in an external Python application
and to debug the application from a new shell inside the Pyzo IDE.
"""


## Example code for any external Python application that you want to debug

# Run this code in a Python >= v3.14 interpreter and remember the PID.
# Or get the PID of any running Python application (>= v3.14) that you want to debug.

import os
import time

def myfunc1():

    def myfunc2():
        cnt2 = 1
        while cnt2 < 5:
            time.sleep(0.2)
            print('cnt2:', cnt2)
            cnt2 += 1

    print('example Python application')
    print('this process has PID', os.getpid())
    cnt1 = 0
    while True:
        cnt1 += 1
        print('cnt1:', cnt1)
        time.sleep(2.0)
        myfunc2()

myfunc1()


##
if False:
##

    # Make sure that the Pyzo IDE is running.

    # Run the following code in another Python >= v3.14 interpreter (e.g. Pyzo shell).

    import inspect
    import sys
    import os

    this_dir = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))
    # or: this_dir = '/path/to/folder/of/start_external_shell_script'

    pid = 14782  # TODO:  <--- change the PID to the one of the external Python application
    sys.remote_exec(pid, os.path.join(this_dir, 'start_external_shell.py'))


    # A few seconds later, a new shell will appear in the Pyzo IDE, and the external
    # Python application is paused.

    # Now you can inspect variables and execute code, like in any other shell in Pyzo.
    # The shell starts with the outermost frame for global and local variables, so
    # you only see "myfunc1" and the imports in the current scope.

    # The code for starting the external shell could be called anywhere where it is
    # convenient for the Python interpreter. For the example code above, it will most
    # likely happen inside myfunc1 or myfunc2.
    # You can switch (from the outermost scope) to the scope of myfunc1 by executing
    # the following line in the shell:
    #     import sys; sys._pyzoInterpreter.switchframe(1)
    # Then you can inspect/modify variable "cnt1".
    # For example, execute "cnt1 = -1234" in the shell.
    # If the start of the external shell happend inside myfunc2, then you can also
    # inspect the values there (cnt2). To switch to that frame, execute the line:
    #     import sys; sys._pyzoInterpreter.switchframe(2)

    # Finally terminate the shell (e.g. via "Shell -> Close" in Pyzo's menu).
    # The external Python application is now running again, but with the modified cnt1 value.

    # You can reconnect again by executing this code cell again.



import os
import sys
import time
import socket
import threading

from yoton.misc import basestring, str
from yoton.misc import getErrorMsg, UID
from yoton.misc import TinyPackageQueue
from yoton.core import Package, PACKAGE_HEARTBEAT, PACKAGE_CLOSE, EINTR
from yoton.core import can_recv, send_all, recv_all, HEADER_SIZE

from yoton.connection import Connection, TIMEOUT_MIN
from yoton.connection import STATUS_CLOSED, STATUS_WAITING, STATUS_HOSTING
from yoton.connection import STATUS_CONNECTED, STATUS_CLOSING  # noqa

# Note that time.sleep(0) yields the current thread's timeslice to any
# other >= priority thread in the process, but is otherwise equivalent 0 delay.


# Reasons to stop the connection
STOP_SOCKET_ERROR = "Socket error."  # the error message is appended
STOP_EOF = "Other end dropped the connection."
STOP_HANDSHAKE_TIMEOUT = "Handshake timed out."
STOP_HANDSHAKE_FAILED = "Handshake failed."
STOP_HANDSHAKE_SELF = "Handshake failed (context cannot connect to self)."
STOP_CLOSED_FROM_THERE = "Closed from other end."
STOP_LOST_TRACK = "Lost track of the stream."
STOP_THREAD_ERROR = "Error in io thread."

# Use a relatively small buffer size, to keep the channels better in sync
SOCKET_BUFFERS_SIZE = 10 * 1024


class TcpConnection(Connection):
    """TcpConnection(context, name='')

    The TcpConnection class implements a connection between two
    contexts that are in different processes or on different machines
    connected via the internet.

    This class handles the low-level communication for the context.
    A ContextConnection instance wraps a single BSD socket for its
    communication, and uses TCP/IP as the underlying communication
    protocol. A persistent connection is used (the BSD sockets stay
    connected). This allows to better distinguish between connection
    problems and timeouts caused by the other side being busy.

    """

    def __init__(self, context, name=""):
        # Variables to hold threads
        self._sendingThread = None
        self._receivingThread = None

        # Create queue for outgoing packages.
        self._qout = TinyPackageQueue(64, *context._queue_params)

        # Init normally (calls _set_status(0)
        Connection.__init__(self, context, name)

    def _set_status(self, status, bsd_socket=None):
        """_connected(status, bsd_socket=None)

        This method is called when a connection is made.

        Private method to apply the bsd_socket.
        Sets the socket and updates the status.
        Also instantiates the IO threads.

        """

        # Lock the connection while we change its status
        with self._lock:
            # Update hostname and port number; for hosting connections the port
            # may be different if max_tries > 0. Each client connection will be
            # assigned a different ephemeral port number.
            # http://www.tcpipguide.com/free/t_TCPPortsConnectionsandConnectionIdentification-2.htm
            # Also get hostname and port for other end
            if bsd_socket is not None:
                if True:
                    self._hostname1, self._port1 = bsd_socket.getsockname()
                if status != STATUS_WAITING:
                    self._hostname2, self._port2 = bsd_socket.getpeername()

            # Set status as normal
            Connection._set_status(self, status)

            if status in [STATUS_HOSTING, STATUS_CONNECTED]:
                # Really connected

                # Store socket
                self._bsd_socket = bsd_socket

                # Set socket to blocking. Needed explicitly on Windows
                # One of these things it takes you hours to find out ...
                bsd_socket.setblocking(1)

                # Create and start io threads
                self._sendingThread = SendingThread(self)
                self._receivingThread = ReceivingThread(self)
                #
                self._sendingThread.start()
                self._receivingThread.start()

            if status == STATUS_CLOSED:
                # Close bsd socket
                try:
                    self._bsd_socket.shutdown()
                except Exception:
                    pass
                try:
                    self._bsd_socket.close()
                except Exception:
                    pass
                self._bsd_socket = None

                # Remove references to threads
                self._sendingThread = None
                self._receivingThread = None

    def _bind(self, hostname, port, max_tries=1):
        """Bind the bsd socket. Launches a dedicated thread that waits
        for incoming connections and to do the handshaking procedure.
        """

        # Create socket.
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Set buffer size to be fairly small (less than 10 packages)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, SOCKET_BUFFERS_SIZE)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, SOCKET_BUFFERS_SIZE)

        # Apply SO_REUSEADDR when binding, so that an improperly closed
        # socket on the same port will not prevent us from connecting.
        # It also allows a connection to bind at the same port number,
        # but only after the previous binding connection has connected
        # (and has closed the listen-socket).
        #
        # SO_REUSEADDR means something different on win32 than it does
        # for Linux sockets. To get the intended behavior on Windows,
        # we don't have to do anything. Also see:
        #  * http://msdn.microsoft.com/en-us/library/ms740621%28VS.85%29.aspx
        #  * http://twistedmatrix.com/trac/ticket/1151
        #  * http://www.tcpipguide.com/free/t_TCPPortsConnectionsandConnectionIdentification-2.htm
        if not sys.platform.startswith("win"):
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # Bind port and try with higher ports if not available
        port_start = port
        for ind_try in range(max_tries):
            try:
                s.bind((hostname, port))
                break
            except Exception:
                # Raise the socket exception if we were asked to try
                # just one port. Otherwise just try the next
                if max_tries == 1:
                    raise
            port += int((ind_try + 1) ** 0.5)
        else:
            # We tried all ports without success
            tmp = "Could not bind to any of the %i ports tried, starting with %i."
            raise IOError(tmp % (max_tries, port_start))

        # Tell the socket it is a host. Backlog of at least 1 to avoid linux
        # kernel from detecting SYN flood and throttling the connection (#381)
        s.listen(1)

        # Set connected (status 1: waiting for connection)
        # Will be called with status 2 by the hostThread on success
        self._set_status(STATUS_WAITING, s)

        # Start thread to wait for a connection
        # (keep reference so the thread-object stays alive)
        self._hostThread = HostThread(self, s)
        self._hostThread.start()

    def _connect(self, hostname, port, timeout=1.0):
        """Connect to a bound socket."""

        # Create socket
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Set buffer size to be fairly small (less than 10 packages)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, SOCKET_BUFFERS_SIZE)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, SOCKET_BUFFERS_SIZE)

        # Refuse ridiculously low timeouts
        if timeout < 0.01:
            timeout = 0.01

        # Try to connect
        ok = False
        timestamp = time.time() + timeout
        while not ok and time.time() < timestamp:
            try:
                s.connect((hostname, port))
                ok = True
                break
            except socket.error:
                pass
            except socket.timeout:
                pass
            time.sleep(timeout / 100.0)

        # Did it work?
        if not ok:
            type, value, tb = sys.exc_info()
            del tb
            err = str(value)
            raise IOError("Cannot connect to %s on %i: %s" % (hostname, port, err))

        # Shake hands
        h = HandShaker(s)
        success, info = h.shake_hands_as_client(self.id1)

        # Problem?
        if not success:
            self._set_status(0)
            if not info:
                info = "problem during handshake"
            raise IOError("Could not connect: " + info)

        # Store id
        self._id2, self._pid2 = info

        # Set connected (status 3: connected as client)
        self._set_status(STATUS_CONNECTED, s)

    def _notify_other_end_of_closing(self):
        """Send PACKAGE_CLOSE."""
        self._qout.push(PACKAGE_CLOSE)
        try:
            self.flush(1.0)
        except Exception:
            pass  # Well, we tried ...

    def _flush(self, timeout=5.0):
        """Put a dummy message on the queue and spinlock until
        the thread has popped it from the queue.
        """

        # Check
        if not self.is_connected:
            RuntimeError("Cannot flush if not connected.")

        # Put dummy package on the queue
        self._qout.push(PACKAGE_HEARTBEAT)

        # Wait for the queue to empty. If it is, the heart beat package
        # may not been send yet, but every package befor it is.
        timestamp = time.time() + timeout
        while self.is_connected and not self._qout.empty():
            time.sleep(0.01)
            if time.time() > timestamp:
                raise RuntimeError("Sending the packages timed out.")

    def _send_package(self, package):
        """Put package on the queue, where the sending thread will
        pick it up.
        """
        self._qout.push(package)

    def _inject_package(self, package):
        """Put package in queue, but bypass potential blocking."""
        self._qout._q.append(package)


class HostThread(threading.Thread):
    """HostThread(context_connection, bds_socket)

    The host thread is used by the ContextConnection when hosting a
    connection. This thread waits for another context to connect
    to it, and then performs the handshaking procedure.

    When a successful connection is made, the context_connection's
    _connected() method is called and this thread then exits.

    """

    def __init__(self, context_connection, bsd_socket):
        threading.Thread.__init__(self)

        # Store connection and socket
        self._context_connection = context_connection
        self._bsd_host_socket = bsd_socket

        # Make daemon
        self.daemon = True

    def run(self):
        """run()

        The main loop. Waits for a connection and performs handshaking
        if successfull.

        """

        # Try making a connection until success or the context is stopped
        while self._context_connection.is_waiting:
            # Wait for connection
            s = self._wait_for_connection()
            if not s:
                continue

            # Check if not closed in the meantime
            if not self._context_connection.is_waiting:
                break

            # Do handshaking
            hs = HandShaker(s)
            success, info = hs.shake_hands_as_host(self._context_connection.id1)
            if success:
                self._context_connection._id2 = info[0]
                self._context_connection._pid2 = info[1]
            else:
                print("Yoton: Handshake failed: " + info)
                continue

            # Success!
            # Close hosting socket, thereby enabling rebinding at the same port
            self._bsd_host_socket.close()
            # Update the status of the connection
            self._context_connection._set_status(STATUS_HOSTING, s)
            # Break out of the loop
            break

        # Remove ref
        del self._context_connection
        del self._bsd_host_socket

    def _wait_for_connection(self):
        """_wait_for_connection()

        The thread will wait here until someone connects. When a
        connection is made, the new socket is returned.

        """

        # Set timeout so that we can check _stop_me from time to time
        self._bsd_host_socket.settimeout(0.25)

        # Wait
        while self._context_connection.is_waiting:
            try:
                s, addr = self._bsd_host_socket.accept()
                return s  # Return the new socket
            except socket.timeout:
                pass
            except socket.error:
                # Skip errors caused by interruptions.
                type, value, tb = sys.exc_info()
                del tb
                if value.errno != EINTR:
                    raise


class HandShaker:
    """HandShaker(bsd_socket)

    Class that performs the handshaking procedure for Tcp connections.

    Essentially, the connecting side starts by sending 'YOTON!'
    followed by its id as a hex string. The hosting side responds
    with the same message (but with a different id).

    This process is very similar to a client/server pattern (both
    messages are also terminated with '\r\n'). This is done such that
    if for example a web client tries to connect, a sensible error
    message can be returned. Or when a ContextConnection tries to connect
    to a web server, it will be able to determine the error gracefully.

    """

    def __init__(self, bsd_socket):
        # Store bsd socket
        self._bsd_socket = bsd_socket

    def shake_hands_as_host(self, id):
        """_shake_hands_as_host(id)

        As the host, we wait for the client to ask stuff, so when
        for example a http client connects, we can stop the connection.

        Returns (success, info), where info is the id of the context at
        the other end, or the error message in case success is False.

        """

        # Make our message with id and pid
        message = "YOTON!%s.%i" % (UID(id).get_hex(), os.getpid())

        # Get request
        request = self._recv_during_handshaking()

        if not request:
            return False, STOP_HANDSHAKE_TIMEOUT
        elif request.startswith("YOTON!"):
            # Get id
            try:
                id2_str, pid2_str = request[6:].split(".", 1)
                id2, pid2 = int(id2_str, 16), int(pid2_str, 10)
            except Exception:
                self._send_during_handshaking("ERROR: could not parse id.")
                return False, STOP_HANDSHAKE_FAILED
            # Respond and return
            self._send_during_handshaking(message)  # returns error?
            if id == id2:
                return False, STOP_HANDSHAKE_SELF
            else:
                return True, (id2, pid2)
        else:
            # Client is not yoton
            self._send_during_handshaking("ERROR: this is Yoton.")
            return False, STOP_HANDSHAKE_FAILED

    def shake_hands_as_client(self, id):
        """_shake_hands_as_client(id)

        As the client, we ask the host whether it is a Yoton context
        and whether the channels we want to support are all right.

        Returns (success, info), where info is the id of the context at
        the other end, or the error message in case success is False.

        """

        # Make our message with id and pid
        message = "YOTON!%s.%i" % (UID(id).get_hex(), os.getpid())

        # Do request
        self._send_during_handshaking(message)  # returns error?

        # Get response
        response = self._recv_during_handshaking()

        # Process
        if not response:
            return False, STOP_HANDSHAKE_TIMEOUT
        elif response.startswith("YOTON!"):
            # Get id
            try:
                id2_str, pid2_str = response[6:].split(".", 1)
                id2, pid2 = int(id2_str, 16), int(pid2_str, 10)
            except Exception:
                return False, STOP_HANDSHAKE_FAILED
            if id == id2:
                return False, STOP_HANDSHAKE_SELF
            else:
                return True, (id2, pid2)
        else:
            return False, STOP_HANDSHAKE_FAILED

    def _send_during_handshaking(self, text, shutdown=False):
        return send_all(self._bsd_socket, text + "\r\n", shutdown)

    def _recv_during_handshaking(self):
        return recv_all(self._bsd_socket, 2.0, True)


class BaseIOThread(threading.Thread):
    """The base class for the sending and receiving IO threads.
    Implements some common functionality.
    """

    def __init__(self, context_connection):
        threading.Thread.__init__(self)

        # Thread will "destruct" when the interpreter shuts down
        self.daemon = True

        # Store (temporarily) the ref to the context connection
        # Also of bsd_socket, because it might be removed before the
        # thread is well up and running.
        self._context_connection = context_connection
        self._bsd_socket = context_connection._bsd_socket

    def run(self):
        """Method to prepare to enter main loop. There is a try-except here
        to catch exceptions caused by interpreter shutdown.
        """

        # Get ref to context connection but make sure the ref
        # if not stored if the thread stops
        context_connection = self._context_connection
        bsd_socket = self._bsd_socket
        del self._context_connection
        del self._bsd_socket

        try:
            # Run and handle exceptions if someting goes wrong
            self.run2(context_connection, bsd_socket)
        except Exception:
            # An error while handling an exception, most probably
            # interpreter shutdown
            pass

    def run2(self, context_connection, bsd_socket):
        """Method to enter main loop. There is a try-except here to
        catch exceptions in the main loop (such as socket errors and
        errors due to bugs in the code.
        """

        # Get classname to use in messages
        className = "yoton." + self.__class__.__name__

        # Define function to write errors
        def writeErr(err):
            sys.__stderr__.write(str(err) + "\n")
            sys.__stderr__.flush()

        try:
            # Enter mainloop
            stop_reason = self._run(context_connection, bsd_socket)

            # Was there a specific reason to stop?
            if stop_reason:
                context_connection.close_on_problem(stop_reason)
            else:
                pass  # Stopped because the connection is gone (already stopped)

        except socket.error:
            # Socket error. Can happen if the other end closed not so nice
            # Do get the socket error message and pass it on.
            msg = STOP_SOCKET_ERROR + getErrorMsg()
            context_connection.close_on_problem("%s, %s" % (className, msg))

        except Exception:
            # Else: woops, an error!
            errmsg = getErrorMsg()
            msg = STOP_THREAD_ERROR + errmsg
            context_connection.close_on_problem("%s, %s" % (className, msg))
            writeErr("Exception in %s." % className)
            writeErr(errmsg)


class SendingThread(BaseIOThread):
    """The thread that reads packages from the queue and sends them over
    the socket. It uses a timeout while reading from the queue, so it can
    send heart beat packages if no packages are sent.
    """

    def _run(self, context_connection, bsd_socket):
        """The main loop. Get package from queue, send package to socket."""

        timeout = 0.5 * TIMEOUT_MIN
        queue = context_connection._qout
        socket_sendall = bsd_socket.sendall

        while True:
            time.sleep(0)  # Be nice

            # Get package from the queue. Use heartbeat package
            # if there have been no packages for a too long time
            try:
                package = queue.pop(timeout)
            except queue.Empty:
                # Use heartbeat package then
                package = PACKAGE_HEARTBEAT
                # Should we stop?
                if not context_connection.is_connected:
                    return None  # No need for a stop message

            # Process the package in parts to avoid data copying (header+data)
            for part in package.parts():
                socket_sendall(part)


class ReceivingThread(BaseIOThread):
    """The thread that reads packages from the socket and passes them to
    the kernel. It uses select() to see if data is available on the socket.
    This allows using a timeout without putting the socket in timeout mode.

    If the timeout has expired, the timedout event for the connection is
    emitted.

    Upon receiving a package, the _recv_package() method of the context
    is called, so this thread will eventually dispose the package in
    one or more queues (of the channel or of another connection).

    """

    def _run(self, context_connection, bsd_socket):
        """The main loop. Get package from socket, deposit package in queue(s)."""

        # Short names in local namespace avoid dictionary lookups
        socket_recv = bsd_socket.recv
        recv_package = context_connection._context._recv_package
        package_from_header = Package.from_header
        HS = HEADER_SIZE

        # Variable to keep track if we emitted a timedout signal
        timedOut = False

        while True:
            time.sleep(0)  # Be nice

            # Use select call on a timeout to test if data is available
            while True:
                try:
                    ok = can_recv(bsd_socket, context_connection._timeout)
                except Exception:
                    # select() has it's share of weird errors
                    raise socket.error("select(): " + getErrorMsg())
                if ok:
                    # Set timedout ex?
                    if timedOut:
                        timedOut = False
                        context_connection.timedout.emit(context_connection, False)
                    # Exit from loop
                    break
                else:
                    # Should we stop?
                    if not context_connection.is_connected:
                        return  # No need for a stop message
                    # Should we do a timeout?
                    if not timedOut:
                        timedOut = True
                        context_connection.timedout.emit(context_connection, True)
                    # Continue in loop
                    continue

            # Get package
            package = self._getPackage(socket_recv, HS, package_from_header)
            if package is None:
                continue
            elif isinstance(package, basestring):
                return package  # error msg

            # Let context handle package further (but happens in this thread)
            try:
                recv_package(package, context_connection)
            except Exception:
                print("Error depositing package in ReceivingThread.")
                print(getErrorMsg())

    def _getPackage(self, socket_recv, HS, package_from_header):
        """Get exactly one package from the socket. Blocking."""

        # Get header and instantiate package object from it
        try:
            header = self._recv_n_bytes(socket_recv, HS)
        except EOFError:
            return STOP_EOF
        package, size = package_from_header(header)

        # Does it look like a good package?
        if not package:
            return STOP_LOST_TRACK

        if size == 0:
            # A special package! (or someone sending a
            # package with no content, which is discarded)
            if package._source_seq == 0:
                pass  # Heart beat
            elif package._source_seq == 1:
                return STOP_CLOSED_FROM_THERE
            return None

        else:
            # Get package data
            try:
                package._data = self._recv_n_bytes(socket_recv, size)
            except EOFError:
                return STOP_EOF
            return package

    def _recv_n_bytes(self, socket_recv, n):
        """Receive exactly n bytes from the socket."""

        # First round
        data = socket_recv(n)
        if len(data) == 0:
            raise EOFError()

        # How many more do we need? For small n, we probably only need 1 round
        n -= len(data)
        if n == 0:
            return data  # We're lucky!

        # Else, we need more than one round
        parts = [data]
        while n:
            data = socket_recv(n)
            parts.append(data)
            n -= len(data)

        # Return combined data of multiple rounds
        return b"".join(parts)



# Import tokens in module namespace
from .python_parser import PythonParser, pythonKeywords

# Set keywords
cythonExtraKeywords = set(
    ["cdef", "cpdef", "ctypedef", "cimport", "float", "double", "int", "long"]
)


class CythonParser(PythonParser):
    """Parser for Cython/Pyrex."""

    _extensions = ["pyi", ".pyx", ".pxd"]

    _keywords = pythonKeywords | cythonExtraKeywords

    def _identifierState(self, identifier=None):
        """Given an identifier returs the identifier state:
        3 means the current identifier can be a function.
        4 means the current identifier can be a class.
        0 otherwise.

        This method enables storing the state during the line,
        and helps the Cython parser to reuse the Python parser's code.

        This implementation keeps a counter. If the counter is 0, the
        state is zero.
        """
        if identifier is None:
            # Explicit get and reset
            state = 0
            try:
                if self._idsCounter > 0:
                    state = self._idsState
            except Exception:
                pass
            self._idsState = 0
            self._idsCounter = 0
            return state
        elif identifier in ["def", "cdef", "cpdef"]:
            # Set function state
            self._idsState = 3
            self._idsCounter = 2
            return 3
        elif identifier == "class":
            # Set class state
            self._idsState = 4
            self._idsCounter = 1
            return 4
        elif self._idsCounter > 0:
            self._idsCounter -= 1
            return self._idsState
        else:
            # This one can be func or class, next one can't
            return 0



"""Module yoton.channels.channels_base

Defines the base channel class and the MessageType class.

"""

import time
import threading

import yoton
from yoton.misc import basestring
from yoton.misc import slot_hash, PackageQueue
from yoton.core import Package
from yoton.context import Context
from yoton.channels.message_types import MessageType, TEXT


class BaseChannel(object):
    """BaseChannel(context, slot_base, message_type=yoton.TEXT)

    Abstract class for all channels.

    Parameters
    ----------
    context : yoton.Context instance
        The context that this channel uses to send messages in a network.
    slot_base : string
        The base slot name. The channel appends an extension to indicate
        message type and messaging pattern to create the final slot name.
        The final slot is used to connect channels at different contexts
        in a network
    message_type : yoton.MessageType instance
        (default is yoton.TEXT)
        Object to convert messages to bytes and bytes to messages.
        Users can create their own message_type class to enable
        communicating any type of message they want.

    Details
    -------
    Messages sent via a channel are delivered asynchronically to the
    corresponding channels.

    All channels are associated with a context and can be used to send
    messages to other channels in the network. Each channel is also
    associated with a slot, which is a string that represents a kind
    of address. A message send by a channel at slot X can only be received
    by a channel with slot X.

    Note that the channel appends an extension
    to the user-supplied slot name, that represents the message type
    and messaging pattern of the channel. In this way, it is prevented
    that for example a PubChannel can communicate with a RepChannel.

    """

    def __init__(self, context, slot_base, message_type=None):
        # Store context
        if not isinstance(context, Context):
            raise ValueError("Context not valid.")
        self._context = context

        # Check message type
        if message_type is None:
            message_type = TEXT
        if isinstance(message_type, type) and issubclass(message_type, MessageType):
            message_type = message_type()
        if isinstance(message_type, MessageType):
            message_type = message_type
        else:
            raise ValueError("message_type should be a MessageType instance.")

        # Store message type and conversion methods
        self._message_type_instance = message_type
        self.message_from_bytes = message_type.message_from_bytes
        self.message_to_bytes = message_type.message_to_bytes

        # Queue for incoming traffic (not used for pure sending channels)
        self._q_in = PackageQueue(*context._queue_params)

        # For sending channels: to lock the channel for sending
        self._send_condition = threading.Condition()
        self._is_send_locked = 0  # "True" is the timeout time

        # Signal for receiving data
        self._received_signal = yoton.events.Signal()
        self._posted_received_event = False

        # Channels can be closed
        self._closed = False

        # Event driven mode
        self._run_mode = 0

        # Init slots
        self._init_slots(slot_base)

    def _init_slots(self, slot_base):
        """Called from __init__ to initialize the slots and perform all checks."""

        # Check if slot is string
        if not isinstance(slot_base, basestring):
            raise ValueError("slot_base must be a string.")

        # Get full slot names, init byte versions
        slots_t = []
        slots_h = []

        # Get extension for message type and messaging pattern
        ext_type = self._message_type_instance.message_type_name()
        ext_patterns = self._messaging_patterns()  # (incoming, outgoing)

        # Normalize and check slot names
        for ext_pattern in ext_patterns:
            if not ext_pattern:
                slots_t.append(None)
                slots_h.append(0)
                continue
            # Get full name
            slot = slot_base + "." + ext_type + "." + ext_pattern
            # Store text version
            slots_t.append(slot)
            # Strip and make lowercase
            slot = slot.strip().lower()
            # Hash
            slots_h.append(slot_hash(slot))

        # Store slots
        self._slot_out = slots_t[0]
        self._slot_in = slots_t[1]
        self._slot_out_h = slots_h[0]
        self._slot_in_h = slots_h[1]

        # Register slots (warn if neither slot is valid)
        if self._slot_out_h:
            self._context._register_sending_channel(
                self, self._slot_out_h, self._slot_out
            )
        if self._slot_in_h:
            self._context._register_receiving_channel(
                self, self._slot_in_h, self._slot_in
            )
        if not self._slot_out_h and not self._slot_in_h:
            raise ValueError("This channel does not have valid slots.")

    def _messaging_patterns(self):
        """Implement to return a string that specifies the pattern
        for sending and receiving, respectively.
        """
        raise NotImplementedError()

    def close(self):
        """Close the channel, i.e. unregisters this channel at the context.

        A closed channel cannot be reused.

        Future attempt to send() messages will result in an IOError
        being raised. Messages currently in the channel's queue can
        still be recv()'ed, but no new messages will be delivered at
        this channel.

        """
        # We keep a reference to the context, otherwise we need locks
        # The context clears the reference to this channel when unregistering.
        self._closed = True
        self._context._unregister_channel(self)

    def _send(self, message, dest_id=0, dest_seq=0):
        """Sends a message of raw bytes without checking whether they're bytes.

        Optionally, dest_id and dest_seq represent the message that
        this message replies to. These are used for the request/reply
        pattern.

        Returns the package that will be sent (or None). The context
        will set _source_id on the package right before
        sending it away.

        """

        # Check if still open
        if self._closed:
            className = self.__class__.__name__
            raise IOError("Cannot send from closed %s %i." % (className, id(self)))

        if message:
            # If send_locked, wait at most one second
            if self._is_send_locked:
                try:
                    with self._send_condition:
                        self._send_condition.wait(1.0)  # wait for notify
                finally:
                    if time.time() > self._is_send_locked:
                        self._is_send_locked = 0
            # Push it on the queue as a package
            slot = self._slot_out_h
            cid = self._context._id
            p = Package(message, slot, cid, 0, dest_id, dest_seq, 0)
            self._context._send_package(p)
            # Return package
            return p
        else:
            return None

    def _recv(self, block):
        """Receive a package (or None)."""

        if block is True:
            # Block for 0.25 seconds so that KeyboardInterrupt works
            while not self._closed:
                try:
                    return self._q_in.pop(0.25)
                except self._q_in.Empty:
                    continue

        else:
            # Block normal
            try:
                return self._q_in.pop(block)
            except self._q_in.Empty:
                return None

    def _set_send_lock(self, value):
        """Set or unset the blocking for the _send() method."""
        # Set send lock variable. We adopt a timeout (10s) just in case
        # the SubChannel that locks the PubChannel gets disconnected and
        # is unable to unlock it.
        if value:
            self._is_send_locked = time.time() + 10.0
        else:
            self._is_send_locked = 0
        # Notify any threads that are waiting in _send()
        if not value:
            with self._send_condition:
                self._send_condition.notify_all()

    ## How packages are inserted in this channel for receiving

    def _inject_package(self, package):
        """Same as _recv_package, but by definition do not block.

        _recv_package is overloaded in SubChannel. _inject_package is not.
        """
        self._q_in.push(package)
        self._maybe_emit_received()

    def _recv_package(self, package):
        """Put package in the queue."""
        self._q_in.push(package)
        self._maybe_emit_received()

    def _maybe_emit_received(self):
        """We want to emit a signal, but in such a way that multiple
        arriving packages result in a single emit. This methods
        only posts an event if it has not been done, or if the previous
        event has been handled.
        """
        if not self._posted_received_event:
            self._posted_received_event = True
            event = yoton.events.Event(self._emit_received)
            yoton.app.post_event(event)

    def _emit_received(self):
        """Emits the "received" signal.

        This method is called once new data has been received.
        However, multiple arrived messages may result in a single call
        to this method. There is also no guarantee that recv() has not
        been called in the meantime.

        Also sets the variable so that a new event for this may be
        created. This method is called from the event loop.

        """
        self._posted_received_event = False  # Reset
        self.received.emit_now(self)

    # Received property sits on the BaseChannel because is is used by almost
    # all channels. Note that PubChannels never emit this signal as they
    # catch status messages from the SubChannel by overloading _recv_package().
    @property
    def received(self):
        """Signal that is emitted when new data is received. Multiple
        arrived messages may result in a single call to this method.
        There is no guarantee that recv() has not been called in the
        meantime. The signal is emitted with the channel instance
        as argument.
        """
        return self._received_signal

    ## Properties

    @property
    def pending(self):
        """Get the number of pending incoming messages."""
        return len(self._q_in)

    @property
    def closed(self):
        """Get whether the channel is closed."""
        return self._closed

    @property
    def slot_outgoing(self):
        """Get the outgoing slot name."""
        return self._slot_out

    @property
    def slot_incoming(self):
        """Get the incoming slot name."""
        return self._slot_in
